from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
import tkinter as tk
import tkinter.filedialog as tkfile
import sys, os
from . import elements

class MainWindow(tk.Tk):
    def __init__(self):
        tk.Tk.__init__(self)

        self.title("ForceBalance 1.1")
        self.geometry("800x600")

        ## Window Components
        # general view of selectable objects in calculation
        self.objectsPane = elements.ObjectViewer(self)
        # more detailed properties relating to selected object(s)
        self.detailsPane = elements.DetailViewer(self)
        # forcebalance output is routed to the console pane
        self.consolePane = elements.ConsoleViewer(self, height = 500)
        
        self.showObjects = tk.IntVar()
        self.showObjects.set(True)
        
        self.arrangeWindow()
        
        self._initialize_menu()

        # Application-wide event bindings
        self.objectsPane.selectionchanged.trace('r', self.updateDetailView)
        self.showObjects.trace('w', self.arrangeWindow)

    def _initialize_menu(self):
        self.menubar = tk.Menu(self)
        filemenu = tk.Menu(self.menubar, tearoff=0)
        filemenu.add_command(label="New", state="disabled")

        filemenu.add_command(label="Open", command=self.open)
        filemenu.add_command(label="Save", state="disabled")
        filemenu.add_command(label="Save As...",state="disabled")
        filemenu.add_command(label="Close", command=self.close)
        filemenu.add_command(label="Exit", command=sys.exit)
        self.menubar.add_cascade(label="File", menu=filemenu)
        
        optionsmenu = tk.Menu(self.menubar, tearoff=0)
        optionsmenu.add_checkbutton(label="show object viewer", variable=self.showObjects)
        optionsmenu.add_checkbutton(label="show default options", variable=self.detailsPane.printAll)
        self.menubar.add_cascade(label="Options", menu=optionsmenu)
        
        self.menubar.add_command(label="Run", command=self.objectsPane.run, state="disabled")

        self['menu']=self.menubar

        #calculationmenu = tk.Menu(self.menubar, tearoff=0)
        #calculationmenu.add_command(label="Check", state="disabled")
        #calculationmenu.add_command(label="Run", command=self.objectsPane.run)
        #self.menubar.add_cascade(label="Calculation", menu=calculationmenu)

    def open(self):
        filters = [('Forcebalance input files', '*.in'),('Show all', '*')]
        inputfile = tkfile.askopenfilename(title="Open ForceBalance input file...", filetypes=filters)
        if inputfile:
            self.consolePane.clear()
            self.objectsPane.open(inputfile)
            self.menubar.entryconfig(3, state=tk.NORMAL)

    def close(self):
        self.objectsPane.clear()
        self.detailsPane.clear()
        self.menubar.entryconfig(3, state=tk.DISABLED)
        
    def arrangeWindow(self, *args):
        self.objectsPane.grid_forget()
        self.detailsPane.grid_forget()
        self.consolePane.grid_forget()
    
        if self.showObjects.get():
            self.consolePane.grid(row=1, columnspan=2, sticky=tk.W+tk.E+tk.N+tk.S)
            self.objectsPane.grid(row=0, column=0, sticky=tk.W+tk.E+tk.N+tk.S)
            self.detailsPane.grid(row=0, column=1, sticky=tk.W+tk.E+tk.N+tk.S)
            self.rowconfigure(0, weight=0)
            self.rowconfigure(1, weight=1)
            self.columnconfigure(0, weight=1)
            self.columnconfigure(1, weight=1)            
            
        else:
            self.consolePane.grid(sticky=tk.W+tk.E+tk.N+tk.S)
            self.rowconfigure(0, weight=1)
            self.rowconfigure(1, weight=0)
            self.columnconfigure(0, weight=1)
            self.columnconfigure(1, weight=0)  

    def updateDetailView(self, *args):
        self.detailsPane.load(self.objectsPane.activeselection)



""" @package forcebalance.abinitio Ab-initio fitting module (energies, forces, resp).

@author Lee-Ping Wang
@date 05/2012
"""
from __future__ import division
from __future__ import print_function

from builtins import zip
from builtins import range
import os
import shutil
from forcebalance.nifty import col, eqcgmx, flat, floatornan, fqcgmx, invert_svd, kb, printcool, bohr2ang, warn_press_key, warn_once, pvec1d, commadash, uncommadash, isint
import numpy as np
from forcebalance.target import Target
from forcebalance.molecule import Molecule, format_xyz_coord
from re import match, sub
import subprocess
from subprocess import PIPE
from forcebalance.finite_difference import fdwrap, f1d2p, f12d3p, in_fd
from collections import defaultdict, OrderedDict
import itertools
#from IPython import embed
#from _increment import AbInitio_Build

from forcebalance.output import getLogger
logger = getLogger(__name__)

class AbInitio(Target):

    """ Subclass of Target for fitting force fields to ab initio data.

    Currently Gromacs-X2, Gromacs, Tinker, OpenMM and AMBER are supported.

    We introduce the following concepts:
    - The number of snapshots
    - The reference energies and forces (eqm, fqm) and the file they belong in (qdata.txt)
    - The sampling simulation energies (emd0)
    - The WHAM Boltzmann weights (these are computed externally and passed in)
    - The QM Boltzmann weights (computed internally using the difference between eqm and emd0)

    There are also these little details:
    - Switches for whether to turn on certain Boltzmann weights (they stack)
    - Temperature for the QM Boltzmann weights
    - Whether to fit a subset of atoms

    This subclass contains the 'get' method for building the objective
    function from any simulation software (a driver to run the program and
    read output is still required).  The 'get' method can be overridden
    by subclasses like AbInitio_GMX."""
    
    def __init__(self,options,tgt_opts,forcefield):
        """
        Initialization; define a few core concepts.

        @todo Obtain the number of true atoms (or the particle -> atom mapping)
        from the force field.
        """

        ## Initialize the base class
        super(AbInitio,self).__init__(options,tgt_opts,forcefield)
        
        #======================================#
        # Options that are given by the parser #
        #======================================#
        
        ## Number of snapshots
        self.set_option(tgt_opts,'shots','ns')
        ## Whether to use WHAM Boltzmann weights
        self.set_option(tgt_opts,'whamboltz','whamboltz')
        ## Whether to use the Sampling Correction
        self.set_option(tgt_opts,'sampcorr','sampcorr')
        ## Whether to match Absolute Energies (make sure you know what you're doing)
        self.set_option(tgt_opts,'absolute','absolute')
        ## Whether to use the Covariance Matrix
        self.set_option(tgt_opts,'covariance','covariance')
        ## Whether to use QM Boltzmann weights
        self.set_option(tgt_opts,'qmboltz','qmboltz')
        ## The temperature for QM Boltzmann weights
        self.set_option(tgt_opts,'qmboltztemp','qmboltztemp')
        ## Number of atoms that we are fitting
        self.set_option(tgt_opts,'fitatoms','fitatoms_in')
        ## Whether to fit Energies.
        self.set_option(tgt_opts,'energy','energy')
        ## Whether to fit Forces.
        self.set_option(tgt_opts,'force','force')
        ## Whether to fit Electrostatic Potential.
        self.set_option(tgt_opts,'resp','resp')
        self.set_option(tgt_opts,'resp_a','resp_a')
        self.set_option(tgt_opts,'resp_b','resp_b')
        ## Weights for the three components.
        self.set_option(tgt_opts,'w_energy','w_energy')
        self.set_option(tgt_opts,'w_force','w_force')
        self.set_option(tgt_opts,'force_map','force_map')
        self.set_option(tgt_opts,'w_netforce','w_netforce')
        self.set_option(tgt_opts,'w_torque','w_torque')
        self.set_option(tgt_opts,'w_resp','w_resp')
        ## Option for how much data to write to disk.
        self.set_option(tgt_opts,'writelevel','writelevel')
        ## Whether to do energy and force calculations for the whole trajectory, or to do
        ## one calculation per snapshot.
        self.set_option(tgt_opts,'all_at_once','all_at_once')
        ## OpenMM-only option - whether to run the energies and forces internally.
        self.set_option(tgt_opts,'run_internal','run_internal')
        ## Whether we have virtual sites (set at the global option level)
        self.set_option(options,'have_vsite','have_vsite')
        ## Attenuate the weights as a function of energy
        self.set_option(tgt_opts,'attenuate','attenuate')
        ## What is the energy denominator? (Valid for 'attenuate')
        self.set_option(tgt_opts,'energy_denom','energy_denom')
        ## Set upper cutoff energy
        self.set_option(tgt_opts,'energy_upper','energy_upper')
        ## Average forces over individual atoms ('atom') or all atoms ('all')
        self.set_option(tgt_opts,'force_average')
        ## Assign a greater weight to MM snapshots that underestimate the QM energy (surfaces referenced to QM absolute minimum)
        self.set_option(tgt_opts,'energy_asymmetry')
        self.savg = (self.energy_asymmetry == 1.0 and not self.absolute)
        self.asym = (self.energy_asymmetry != 1.0)
        if self.asym:
            if not self.all_at_once:
                logger.error("Asymmetric weights only work when all_at_once is enabled")
                raise RuntimeError
            if self.qmboltz != 0.0:
                logger.error("Asymmetric weights do not work with QM Boltzmann weights")
                raise RuntimeError
        #======================================#
        #     Variables which are set here     #
        #======================================#
        ## WHAM Boltzmann weights
        self.boltz_wts = []
        ## QM Boltzmann weights
        self.qmboltz_wts   = []
        ## Reference (QM) energies
        self.eqm           = []
        ## Energies of the sampling simulation
        self.emd0          = []
        ## Reference (QM) forces
        self.fqm           = []
        ## ESP grid points
        self.espxyz        = []
        ## ESP values
        self.espval        = []
        ## The qdata.txt file that contains the QM energies and forces
        self.qfnm = os.path.join(self.tgtdir,"qdata.txt")
        ## Qualitative Indicator: average energy error (in kJ/mol)
        self.e_err = 0.0
        self.e_err_pct = None
        ## Qualitative Indicator: average force error (fractional)
        self.f_err = 0.0
        self.f_err_pct = 0.0
        ## Qualitative Indicator: "relative RMS" for electrostatic potential
        self.esp_err = 0.0
        self.nf_err = 0.0
        self.nf_err_pct = 0.0
        self.tq_err_pct = 0.0
        ## Whether to compute net forces and torques, or not.
        self.use_nft       = self.w_netforce > 0 or self.w_torque > 0
        ## Read in the trajectory file
        self.mol = Molecule(os.path.join(self.root,self.tgtdir,self.coords), 
                            top=(os.path.join(self.root,self.tgtdir,self.pdb) if hasattr(self, 'pdb') else None))
        ## Set the number of snapshots
        if self.ns != -1:
            self.mol = self.mol[:self.ns]
        self.ns = len(self.mol)
        ## The number of (atoms + drude particles + virtual sites)
        self.nparticles  = len(self.mol.elem)
        ## Build keyword dictionaries to pass to engine.
        engine_args = OrderedDict(list(self.OptionDict.items()) + list(options.items()))
        del engine_args['name']
        ## Create engine object.
        self.engine = self.engine_(target=self, mol=self.mol, **engine_args)
        ## Lists of atoms to do net force/torque fitting and excluding virtual sites.
        self.AtomLists = self.engine.AtomLists
        self.AtomMask = self.engine.AtomMask
        ## Read in the reference data
        self.read_reference_data()
        ## The below two options are related to whether we want to rebuild virtual site positions.
        ## Rebuild the distance matrix if virtual site positions have changed
        self.buildx = True
        ## Save the mvals from the last time we updated the vsites.
        self.save_vmvals = {}
        self.set_option(None, 'shots', val=self.ns)

    def build_invdist(self, mvals):
        for i in self.pgrad:
            if 'VSITE' in self.FF.plist[i]:
                if i in self.save_vmvals and mvals[i] != self.save_vmvals[i]:
                    self.buildx = True
                    break
        if not self.buildx: return self.invdists
        if any(['VSITE' in i for i in self.FF.map.keys()]) or self.have_vsite:
            logger.info("\rGenerating virtual site positions.%s" % (" "*30))
            pvals = self.FF.make(mvals)
            self.mol.xyzs = self.engine.generate_positions()
        # prepare the distance matrix for esp computations
        if len(self.espxyz) > 0:
            invdists = []
            logger.info("\rPreparing the distance matrix... it will have %i * %i * %i = %i elements" % (self.ns, self.nesp, self.nparticles, self.ns * self.nesp * self.nparticles))
            sn = 0
            for espset, xyz in zip(self.espxyz, self.mol.xyzs):
                logger.info("\rGenerating ESP distances for snapshot %i%s\r" % (sn, " "*50))
                esparr = np.array(espset).reshape(-1,3)
                # Create a matrix with Nesp rows and Natoms columns.
                DistMat = np.array([[np.linalg.norm(i - j) for j in xyz] for i in esparr])
                invdists.append(1. / (DistMat / bohr2ang))
                sn += 1
        for i in self.pgrad:
            if 'VSITE' in self.FF.plist[i]:
                self.save_vmvals[i] = mvals[i]
        self.buildx = False
        return np.array(invdists)

    def compute_netforce_torque(self, xyz, force, QM=False):
        # Convert an array of (3 * n_atoms) atomistic forces
        # to an array of (3 * (n_forces + n_torques)) net forces and torques.
        # This code is rather slow.  It requires the system to have a list
        # of masses and blocking numbers.
       
        kwds = {"MoleculeNumber" : "molecule",
                "ResidueNumber" : "residue",
                "ChargeGroupNumber" : "chargegroup"}
        if self.force_map == 'molecule' and 'MoleculeNumber' in self.AtomLists:
            Block = self.AtomLists['MoleculeNumber']
        elif self.force_map == 'residue' and 'ResidueNumber' in self.AtomLists:
            Block = self.AtomLists['ResidueNumber']
        elif self.force_map == 'chargegroup' and 'ChargeGroupNumber' in self.AtomLists:
            Block = self.AtomLists['ChargeGroupNumber']
        else:
            logger.error('force_map keyword "%s" is invalid. Please choose from: %s\n' % (self.force_map, ', '.join(['"%s"' % kwds[k] for k in self.AtomLists.keys() if k in kwds])))
            raise RuntimeError

        nft = len(self.fitatoms)
        # Number of particles that the force is acting on
        nfp = force.reshape(-1,3).shape[0]
        # Number of particles in the XYZ coordinates
        nxp = xyz.shape[0]
        # Number of particles in self.AtomLists
        npr = len(self.AtomMask)
        # Number of true atoms
        nat = sum(self.AtomMask)

        mask = np.array([i for i in range(npr) if self.AtomMask[i]])
        
        if nfp not in [npr, nat]:
            logger.error('Force contains %i particles but expected %i or %i\n' % (nfp, npr, nat))
            raise RuntimeError
        elif nfp == nat:
            frc1 = force.reshape(-1,3)[:nft].flatten()
        elif nfp == npr:
            frc1 = force.reshape(-1,3)[mask][:nft].flatten()

        if nxp not in [npr, nat]:
            logger.error('Coordinates contains %i particles but expected %i or %i\n' % (nfp, npr, nat))
            raise RuntimeError
        elif nxp == nat:
            xyz1 = xyz[:nft]
        elif nxp == npr:
            xyz1 = xyz[mask][:nft]

        Block = list(np.array(Block)[mask])[:nft]
        Mass = np.array(self.AtomLists['Mass'])[mask][:nft]

        NetForces = []
        Torques = []
        for b in sorted(set(Block)):
            AtomBlock = np.array([i for i in range(len(Block)) if Block[i] == b])
            CrdBlock = np.array(list(itertools.chain(*[list(range(3*i, 3*i+3)) for i in AtomBlock])))
            com = np.sum(xyz1[AtomBlock]*np.outer(Mass[AtomBlock],np.ones(3)), axis=0) / np.sum(Mass[AtomBlock])
            frc = frc1[CrdBlock].reshape(-1,3)
            NetForce = np.sum(frc, axis=0)
            xyzb = xyz1[AtomBlock]
            Torque = np.zeros(3)
            for a in range(len(xyzb)):
                R = xyzb[a] - com
                F = frc[a]
                # I think the unit of torque is in nm x kJ / nm.
                Torque += np.cross(R, F) / 10
            NetForces += [i for i in NetForce]
            # Increment the torques only if we have more than one atom in our block.
            if len(xyzb) > 1:
                Torques += [i for i in Torque]
        netfrc_torque = np.array(NetForces + Torques)
        self.nnf = int(len(NetForces)/3)
        self.ntq = int(len(Torques)/3)
        return netfrc_torque

    def read_reference_data(self):
        
        """ Read the reference ab initio data from a file such as qdata.txt.

        @todo Add an option for picking any slice out of
        qdata.txt, helpful for cross-validation
        
        @todo Closer integration of reference data with program -
        leave behind the qdata.txt format?  (For now, I like the
        readability of qdata.txt)

        After reading in the information from qdata.txt, it is converted
        into the GROMACS energy units (kind of an arbitrary choice);
        forces (kind of a misnomer in qdata.txt) are multipled by -1
        to convert gradients to forces.

        We also subtract out the mean energies of all energy arrays
        because energy/force matching does not account for zero-point
        energy differences between MM and QM (i.e. energy of electrons
        in core orbitals).

        The configurations in force/energy matching typically come
        from a the thermodynamic ensemble of the MM force field at
        some temperature (by running MD, for example), and for many
        reasons it is helpful to introduce non-Boltzmann weights in
        front of these configurations.  There are two options: WHAM
        Boltzmann weights (for combining the weights of several
        simulations together) and QM Boltzmann weights (for converting
        MM weights into QM weights).  Note that the two sets of weights
        'stack'; i.e. they can be used at the same time.

        A 'hybrid' ensemble is possible where we use 50% MM and 50% QM
        weights.  Please read more in LPW and Troy Van Voorhis, JCP
        Vol. 133, Pg. 231101 (2010), doi:10.1063/1.3519043.
        
        @todo The WHAM Boltzmann weights are generated by external
        scripts (wanalyze.py and make-wham-data.sh) and passed in;
        perhaps these scripts can be added to the ForceBalance
        distribution or integrated more tightly.

        Finally, note that using non-Boltzmann weights degrades the
        statistical information content of the snapshots.  This
        problem will generally become worse if the ensemble to which
        we're reweighting is dramatically different from the one we're
        sampling from.  We end up with a set of Boltzmann weights like
        [1e-9, 1e-9, 1.0, 1e-9, 1e-9 ... ] and this is essentially just
        one snapshot.  I believe Troy is working on something to cure
        this problem.

        Here, we have a measure for the information content of our snapshots,
        which comes easily from the definition of information entropy:

        S = -1*Sum_i(P_i*log(P_i))
        InfoContent = exp(-S)

        With uniform weights, InfoContent is equal to the number of snapshots;
        with horrible weights, InfoContent is closer to one.

        """
        # Parse the qdata.txt file
        for line in open(os.path.join(self.root,self.qfnm)):
            sline = line.split()
            if len(sline) == 0: continue
            elif sline[0] == 'ENERGY':
                self.eqm.append(float(sline[1]))
            elif sline[0] == 'EMD0':
                self.emd0.append(float(sline[1]))
            elif sline[0] == 'FORCES':
                self.fqm.append([float(i) for i in sline[1:]])
            elif sline[0] == 'ESPXYZ':
                self.espxyz.append([float(i) for i in sline[1:]])
            elif sline[0] == 'ESPVAL':
                self.espval.append([float(i) for i in sline[1:]])
        
        # Ensure that all lists are of length self.ns
        self.eqm = self.eqm[:self.ns]
        self.emd0 = self.emd0[:self.ns]
        self.fqm = self.fqm[:self.ns]
        self.espxyz = self.espxyz[:self.ns]
        self.espval = self.espval[:self.ns]

        # Turn everything into arrays, convert to kJ/mol, and subtract the mean energy from the energy arrays
        self.eqm = np.array(self.eqm)
        self.eqm *= eqcgmx
        if self.asym:
            self.eqm  -= np.min(self.eqm)
            self.smin  = np.argmin(self.eqm)
            logger.info("Referencing all energies to the snapshot %i (minimum energy structure in QM)\n" % self.smin)
        elif self.absolute:
            logger.info("Fitting absolute energies.  Make sure you know what you are doing!\n")
        else:
            self.eqm -= np.mean(self.eqm)

        if len(self.fqm) > 0:
            self.fqm = np.array(self.fqm)
            self.fqm *= fqcgmx
            self.qmatoms = list(range(int(self.fqm.shape[1]/3)))
        else:
            logger.info("QM forces are not present, only fitting energies.\n")
            self.force = 0
            self.w_force = 0

        # Here we may choose a subset of atoms to do the force matching.
        if self.force:
            # Build a list corresponding to the atom indices where we are fitting the forces.
            if isint(self.fitatoms_in):
                if int(self.fitatoms_in) == 0:
                    self.fitatoms = self.qmatoms
                else:
                    warn_press_key("Provided an integer for fitatoms; will assume this means the first %i atoms" % int(self.fitatoms_in))
                    self.fitatoms = list(range(int(self.fitatoms_in)))
            else:
                # If provided a "comma and dash" list, then expand the list.
                self.fitatoms = uncommadash(self.fitatoms_in)

            if len(self.fitatoms) > len(self.qmatoms):
                warn_press_key("There are more fitting atoms than the total number of atoms in the QM calculation (something is probably wrong)")
            else:
                if self.w_force > 0:
                    if len(self.fitatoms) == len(self.qmatoms):
                        logger.info("Fitting the forces on all atoms\n")
                    else:
                        logger.info("Fitting the forces on atoms %s\n" % commadash(self.fitatoms))
                        logger.info("Pruning the quantum force matrix...\n")
                selct = list(itertools.chain(*[[3*i+j for j in range(3)] for i in self.fitatoms]))
                self.fqm  = self.fqm[:, selct]
        else:
            self.fitatoms = []

        self.nesp = len(self.espval[0]) if len(self.espval) > 0 else 0
            
        if len(self.emd0) > 0:
            self.emd0 = np.array(self.emd0)
            self.emd0 -= np.mean(self.emd0)

        if self.whamboltz == True:
            if self.attenuate:
                logger.error('whamboltz and attenuate are mutually exclusive\n')
                raise RuntimeError
            self.boltz_wts = np.array([float(i.strip()) for i in open(os.path.join(self.root,self.tgtdir,"wham-weights.txt")).readlines()])
            #   This is a constant pre-multiplier in front of every snapshot.
            bar = printcool("Using WHAM MM Boltzmann weights.", color=4)
            if os.path.exists(os.path.join(self.root,self.tgtdir,"wham-master.txt")):
                whaminfo = open(os.path.join(self.root,self.tgtdir,"wham-master.txt")).readlines()
                logger.info("From wham-master.txt, I can see that you're using %i generations\n" % len(whaminfo))
                logger.info("Relative weight of each generation:\n")
                shotcounter = 0
                for line in whaminfo:
                    sline = line.split()
                    genshots = int(sline[2])
                    weight = np.sum(self.boltz_wts[shotcounter:shotcounter+genshots])/np.sum(self.boltz_wts)
                    logger.info(" %s, %i snapshots, weight %.3e\n" % (sline[0], genshots, weight))
                    shotcounter += genshots
            else:
                logger.info("Oops... WHAM files don't exist?\n")
            logger.info(bar)
        elif self.attenuate:
            # Attenuate energies by an amount proportional to their
            # value above the minimum.
            eqm1 = self.eqm - np.min(self.eqm)
            denom = self.energy_denom * 4.184 # kcal/mol to kJ/mol
            upper = self.energy_upper * 4.184 # kcal/mol to kJ/mol
            self.boltz_wts = np.ones(self.ns)
            for i in range(self.ns):
                if eqm1[i] > upper:
                    self.boltz_wts[i] = 0.0
                elif eqm1[i] < denom:
                    self.boltz_wts[i] = 1.0 / denom
                else:
                    self.boltz_wts[i] = 1.0 / np.sqrt(denom**2 + (eqm1[i]-denom)**2)
        else:
            self.boltz_wts = np.ones(self.ns)
        
        if self.qmboltz > 0:
            # LPW I haven't revised this section yet
            # Do I need to?
            logboltz = -(self.eqm - np.mean(self.eqm) - self.emd0 + np.mean(self.emd0)) / kb / self.qmboltztemp
            logboltz -= np.max(logboltz) # Normalizes boltzmann weights
            self.qmboltz_wts = np.exp(logboltz)
            self.qmboltz_wts /= np.sum(self.qmboltz_wts)
            # where simply gathers the nonzero values otherwise we get lots of NaNs
            qbwhere = self.qmboltz_wts[np.where(self.qmboltz_wts)[0]]
            # Compute ze InfoContent!
            qmboltzent = -np.sum(qbwhere*np.log(qbwhere))
            logger.info("Quantum Boltzmann weights are ON, the formula is exp(-b(E_qm-E_mm)),")
            logger.info("distribution entropy is %.3f, equivalent to %.2f snapshots\n" % (qmboltzent, np.exp(qmboltzent)))
            logger.info("%.1f%% is mixed into the MM boltzmann weights.\n" % (self.qmboltz*100))
        else:
            self.qmboltz_wts = np.ones(self.ns)
        # At this point, self.fqm is a (number of snapshots) x (3 x number of atoms) array.
        # Now we can transform it into a (number of snapshots) x (3 x number of residues + 3 x number of residues) array.
        if self.use_nft:
            self.nftqm = []
            for i in range(len(self.fqm)):
                self.nftqm.append(self.compute_netforce_torque(self.mol.xyzs[i], self.fqm[i]))
            self.nftqm = np.array(self.nftqm)
            self.fref = np.hstack((self.fqm, self.nftqm))
        else:
            self.fref = self.fqm
            self.nnf = 0
            self.ntq = 0

        # Normalize Boltzmann weights.
        self.boltz_wts /= sum(self.boltz_wts)
        self.qmboltz_wts /= sum(self.qmboltz_wts)

    def indicate(self):
        Headings = ["Observable", "Difference\n(Calc-Ref)", "Denominator\n RMS (Ref)", " Percent \nDifference", "Weight", "Contribution"]
        Data = OrderedDict([])
        if self.energy:
            Data['Energy (kJ/mol)'] = ["%8.4f" % self.e_err,
                                       "%8.4f" % self.e_ref,
                                       "%.4f%%" % (self.e_err_pct*100),
                                       "%.3f" % self.w_energy,
                                       "%8.4f" % self.e_ctr]
        if self.force:
            Data['Gradient (kJ/mol/A)'] = ["%8.4f" % (self.f_err/10),
                                           "%8.4f" % (self.f_ref/10),
                                           "%.4f%%" % (self.f_err_pct*100),
                                           "%.3f" % self.w_force,
                                           "%8.4f" % self.f_ctr]
            if self.use_nft:
                Data['Net Force (kJ/mol/A)'] = ["%8.4f" % (self.nf_err/10),
                                                "%8.4f" % (self.nf_ref/10),
                                                "%.4f%%" % (self.nf_err_pct*100),
                                                "%.3f" % self.w_netforce,
                                                "%8.4f" % self.nf_ctr]
                Data['Torque (kJ/mol/rad)'] = ["%8.4f" % self.tq_err,
                                               "%8.4f" % self.tq_ref,
                                               "%.4f%%" % (self.tq_err_pct*100),
                                               "%.3f" % self.w_torque,
                                               "%8.4f" % self.tq_ctr]
        if self.resp:
            Data['Potential (a.u.'] = ["%8.4f" % (self.esp_err/10),
                                       "%8.4f" % (self.esp_ref/10),
                                       "%.4f%%" % (self.esp_err_pct*100),
                                       "%.3f" % self.w_resp,
                                       "%8.4f" % self.esp_ctr]
        self.printcool_table(data=Data, headings=Headings, color=0)
        if self.force:
            logger.info("Maximum force error on atom %i (%s), frame %i, %8.4f kJ/mol/A\n" % (self.maxfatom, self.mol.elem[self.fitatoms[self.maxfatom]], self.maxfshot, self.maxdf/10))

    def energy_all(self):
        if hasattr(self, 'engine'):
            return self.engine.energy().reshape(-1,1)
        else:
            logger.error("Target must contain an engine object\n")
            raise NotImplementedError

    def energy_force_all(self):
        if hasattr(self, 'engine'):
            return self.engine.energy_force()
        else:
            logger.error("Target must contain an engine object\n")
            raise NotImplementedError
        
    def energy_force_transform(self):
        if self.force:
            M = self.energy_force_all()
            selct = [0] + list(itertools.chain(*[[1+3*i+j for j in range(3)] for i in self.fitatoms]))
            M = M[:, selct]
            if self.use_nft:
                Nfts = []
                for i in range(len(M)):
                    Fm  = M[i][1:]
                    Nft = self.compute_netforce_torque(self.mol.xyzs[i], Fm)
                    Nfts.append(Nft)
                Nfts = np.array(Nfts)
                return np.hstack((M, Nfts))
            else:
                return M
        else:
            return self.energy_all()

    def energy_one(self, i):
        if hasattr(self, 'engine'):
            return self.engine.energy_one(i)
        else:
            logger.error("Target must contain an engine object\n")
            raise NotImplementedError

    def energy_force_one(self, i):
        if hasattr(self, 'engine'):
            return self.engine.energy_force_one(i)
        else:
            logger.error("Target must contain an engine object\n")
            raise NotImplementedError

    def energy_force_transform_one(self,i):
        if self.force:
            M = self.energy_force_one(i)
            selct = [0] + list(itertools.chain(*[[1+3*i+j for j in range(3)] for i in self.fitatoms]))
            M = M[:, selct]
            if self.use_nft:
                Fm  = M[1:]
                Nft = self.compute_netforce_torque(self.mol.xyzs[i], Fm)
                return np.hstack((M, Nft))
            else:
                return M
        else:
            return self.energy_one()

    def get_energy_force(self, mvals, AGrad=False, AHess=False):
        """
        LPW 06-30-2013
        
        This subroutine builds the objective function (and optionally
        its derivatives) from a general simulation software.  This is
        in contrast to using GROMACS-X2, which computes the objective
        function and prints it out; then 'get' only needs to call
        GROMACS and read it in.

        This subroutine interfaces with simulation software 'drivers'.
        The driver is only expected to give the energy and forces.

        Now this subroutine may sound trivial since the objective
        function is simply a least-squares quantity (M-Q)^2 - but
        there are a number of nontrivial considerations.  I will list
        them here.
        
        0) Polytensor formulation: Because there may exist covariance
        between different components of the force (or covariance
        between the energy and the force), we build the objective
        function by taking outer products of vectors that have the
        form [E F_1x F_1y F_1z F_2x F_2y ... ], and then we trace it
        with the inverse of the covariance matrix to get the objective
        function.

        This version implements both the polytensor formulation and the standard formulation.

        1) Boltzmann weights and normalization: Each snapshot has its
        own Boltzmann weight, which may or may not be normalized.
        This subroutine does the normalization automatically.
        
        2) Subtracting out the mean energy gap: The zero-point energy
        difference between reference data and simulation is
        meaningless.  This subroutine subtracts it out.
        
        3) Hybrid ensembles: This program builds a combined objective
        function from both MM and QM ensembles, which is rigorously
        better than using a single ensemble.

        Note that this subroutine does not do EVERYTHING that
        GROMACS-X2 can do, which includes:
        
        1) Internal coordinate systems
        2) 'Sampling correction' (deprecated, since it doesn't seem to work)
        3) Analytic derivatives
        
        In the previous code (ForTune) this subroutine used analytic
        first derivatives of the energy and force to build the
        derivatives of the objective function.  Here I will take a
        simplified approach, because building the derivatives are
        cumbersome.  For now we will return the objective function
        ONLY.  A two-point central difference should give us the first
        and diagonal second derivative anyhow.

        @todo Parallelization over snapshots is not implemented yet

        @param[in] mvals Mathematical parameter values
        @param[in] AGrad Switch to turn on analytic gradient
        @param[in] AHess Switch to turn on analytic Hessian
        @return Answer Contribution to the objective function
        """
        cv = self.covariance # Whether covariance matrix is turned on
        if cv and (AGrad or AHess):
            warn_press_key("Covariance is turned on, gradients requested but not implemented (will skip over grad.)")
        Answer = {}
        # Create the new force field!!
        pvals = self.FF.make(mvals)

        #======================================#
        #   Copied from the old ForTune code   #
        #======================================#
        nat  = len(self.fitatoms)
        nnf  = self.nnf
        ntq  = self.ntq
        NC   = 3*nat
        NCP1 = 3*nat+1
        if self.use_nft:
            NCP1 += 3*(nnf + ntq)
        NP   = self.FF.np
        NS   = self.ns
        #==============================================================#
        # Note: Because of hybrid ensembles, we form two separate      #
        # objective functions.  This means the code is (trivially)     #
        # doubled in length because there are two sets of Boltzmann    #
        # weights.  In principle I could write a general subroutine    #
        # for a single set of Boltzmann weights and call it twice, but #
        # that would require me to loop over the snapshots twice.      #
        # However, after the loop over snapshots, the subroutine that  #
        # builds the objective function (build_objective, above) is    #
        # called twice using the MM-ensemble and QM-ensemble           #
        # variables.                                                   #
        #                                                              #
        #            STEP 1: Form all of the arrays.                   #
        #                                                              #
        # An explanation of variable names follows.                    #
        # Z(Y)  = The partition function in the MM (QM) ensemble       #
        # M0_M  = The mean of the MM values in the MM ensemble         #
        # QQ_Q  = The unnormalized expected value of <Q(X)Q>           #
        # Later on we divide these quantities by Z to normalize,       #
        # and optionally Q0(X)Q0 is subtracted from QQ to get the      #
        # covariance.                                                  #
        #==============================================================#
        if (self.w_energy == 0.0 and self.w_force == 0.0 and self.w_netforce == 0.0 and self.w_torque == 0.0):
            AGrad = False
            AHess = False
        Z       = 0.0
        Y       = 0.0
        # The average force / net force per atom in kJ/mol/nm. (Added LPW 6/30)
        # Torques are in kj/mol/nm x nm.
        qF_M     = 0
        qN_M     = 0
        qT_M     = 0
        qF_Q     = 0
        qN_Q     = 0
        qT_Q     = 0
        dF_M     = 0
        dN_M     = 0
        dT_M     = 0
        dF_Q     = 0
        dN_Q     = 0
        dT_Q     = 0
        Q = np.zeros(NCP1)
        # Mean quantities
        M0_M    = np.zeros(NCP1)
        Q0_M    = np.zeros(NCP1)
        M0_Q    = np.zeros(NCP1)
        Q0_Q    = np.zeros(NCP1)
        if cv:
            QQ_M    = np.zeros((NCP1,NCP1))
            QQ_Q    = np.zeros((NCP1,NCP1))
            #==============================================================#
            # Objective function polytensors: This is formed in a loop     #
            # over snapshots by taking the outer product (Q-M)(X)(Q-M),    #
            # multiplying by the Boltzmann weight, and then summing.       #
            #==============================================================#
            SPiXi = np.zeros((NCP1,NCP1))
            SRiXi = np.zeros((NCP1,NCP1))
        else:
            # Derivatives
            M_p     = np.zeros((NP,NCP1))
            M_pp    = np.zeros((NP,NCP1))
            X0_M    = np.zeros(NCP1)
            QQ_M    = np.zeros(NCP1)
            X0_Q    = np.zeros(NCP1)
            Q0_Q    = np.zeros(NCP1)
            QQ_Q    = np.zeros(NCP1)
            # Means of gradients
            M0_M_p  = np.zeros((NP,NCP1))
            M0_Q_p  = np.zeros((NP,NCP1))
            M0_M_pp = np.zeros((NP,NCP1))
            M0_Q_pp = np.zeros((NP,NCP1))
            # Objective functions
            SPiXi = np.zeros(NCP1)
            SRiXi = np.zeros(NCP1)
            # Debug: Store all objective function contributions
            XiAll = np.zeros((NS, NCP1))
            if AGrad:
                SPiXi_p = np.zeros((NP,NCP1))
                SRiXi_p = np.zeros((NP,NCP1))
                X2_M_p = np.zeros(NP)
                X2_Q_p = np.zeros(NP)
            if AHess:
                SPiXi_pq = np.zeros((NP,NP,NCP1))
                SRiXi_pq = np.zeros((NP,NP,NCP1))
                X2_M_pq = np.zeros((NP,NP))
                X2_Q_pq = np.zeros((NP,NP))
            M_all = np.zeros((NS,NCP1))
            if AGrad and self.all_at_once:
                dM_all = np.zeros((NS,NP,NCP1))
                ddM_all = np.zeros((NS,NP,NCP1))
        QBN = np.dot(self.qmboltz_wts[:NS],self.boltz_wts[:NS])
        #==============================================================#
        #             STEP 2: Loop through the snapshots.              #
        #==============================================================#
        if self.all_at_once:
            logger.debug("\rExecuting\r")
            M_all = self.energy_force_transform()
            if self.asym:
                M_all[:, 0] -= M_all[self.smin, 0]
            if not cv and (AGrad or AHess):
                def callM(mvals_):
                    logger.debug("\r")
                    pvals = self.FF.make(mvals_)
                    return self.energy_force_transform()
                for p in self.pgrad:
                    dM_all[:,p,:], ddM_all[:,p,:] = f12d3p(fdwrap(callM, mvals, p), h = self.h, f0 = M_all)
                    if self.asym:
                        dM_all[:, p, 0] -= dM_all[self.smin, p, 0]
                        ddM_all[:, p, 0] -= ddM_all[self.smin, p, 0]
        if self.force and not in_fd():
            self.maxfatom = -1
            self.maxfshot = -1
            self.maxdf = 0.0
        for i in range(NS):
            if i % 100 == 0:
                logger.debug("\rIncrementing quantities for snapshot %i\r" % i)
            # Build Boltzmann weights and increment partition function.
            P   = self.boltz_wts[i]
            Z  += P
            R   = self.qmboltz_wts[i]*self.boltz_wts[i] / QBN
            Y  += R
            # Recall reference (QM) data
            Q[0] = self.eqm[i]
            if self.force:
                Q[1:] = self.fref[i,:].copy()
            # Increment the average quantities.
            if cv:
                QQ     = np.outer(Q,Q)
            else:
                QQ     = Q*Q
            # Call the simulation software to get the MM quantities.
            if self.all_at_once:
                M = M_all[i]
            else:
                if i % 100 == 0:
                    logger.debug("Shot %i\r" % i)
                M = self.energy_force_transform_one(i)
                M_all[i,:] = M.copy()
            if not cv:
                X     = M-Q
                boost   = 1.0
                if self.asym and X[0] < 0.0:
                    boost = self.energy_asymmetry
            # Increment the average values.
            a = 1
            if self.force:
                dfrcarray_ = np.array([np.linalg.norm(M[a+3*j:a+3*j+3] - Q[a+3*j:a+3*j+3]) for j in range(nat)])
                if not in_fd() and np.max(dfrcarray_) > self.maxdf:
                    self.maxdf = np.max(dfrcarray_)
                    self.maxfatom = np.argmax(dfrcarray_)
                    self.maxfshot = i
                dfrcarray = np.mean(dfrcarray_)
                qfrcarray = np.mean(np.array([np.linalg.norm(Q[a+3*j:a+3*j+3]) for j in range(nat)]))
                dF_M    += P*dfrcarray
                dF_Q    += R*dfrcarray
                qF_M    += P*qfrcarray
                qF_Q    += R*qfrcarray
                a       += 3*nat
                if self.use_nft:
                    dnfrcarray = np.mean(np.array([np.linalg.norm(M[a+3*j:a+3*j+3] - Q[a+3*j:a+3*j+3]) for j in range(nnf)]))
                    qnfrcarray = np.mean(np.array([np.linalg.norm(Q[a+3*j:a+3*j+3]) for j in range(nnf)]))
                    dN_M    += P*dnfrcarray
                    dN_Q    += R*dnfrcarray
                    qN_M    += P*qnfrcarray
                    qN_Q    += R*qnfrcarray
                    a       += 3*nnf
                    dtrqarray = np.mean(np.array([np.linalg.norm(M[a+3*j:a+3*j+3] - Q[a+3*j:a+3*j+3]) for j in range(ntq)]))
                    qtrqarray = np.mean(np.array([np.linalg.norm(Q[a+3*j:a+3*j+3]) for j in range(ntq)]))
                    dT_M    += P*dtrqarray
                    dT_Q    += R*dtrqarray
                    qT_M    += P*qtrqarray
                    qT_Q    += R*qtrqarray
            # The [0] indicates that we are fitting the RMS force and not the RMSD
            # (without the covariance, subtracting a mean force doesn't make sense.)
            M0_M[0] += P*M[0]
            M0_Q[0] += R*M[0]
            Q0_M[0] += P*Q[0]
            Q0_Q[0] += R*Q[0]
            QQ_M += P*QQ
            QQ_Q += R*QQ
            if not cv:
                X0_M[0] += P*X[0]
                X0_Q[0] += R*X[0]
            # Increment the objective function.
            if cv:
                Xi  = np.outer(M,M) - 2*np.outer(Q,M) + np.outer(Q,Q)
            else:
                Xi     = X**2                   
                Xi[0] *= boost
            XiAll[i] = Xi.copy()
            SPiXi += P * Xi
            SRiXi += R * Xi
            #==============================================================#
            #      STEP 2a: Increment gradients and mean quantities.       #
            #   This is only implemented for the case without covariance.  #
            #==============================================================#
            if not cv:
                for p in self.pgrad:
                    if not AGrad: continue
                    if self.all_at_once:
                        M_p[p] = dM_all[i, p]
                        M_pp[p] = ddM_all[i, p]
                    else:
                        def callM(mvals_):
                            if i % 100 == 0:
                                logger.debug("\r")
                            pvals = self.FF.make(mvals_)
                            return self.energy_force_transform_one(i)
                        M_p[p],M_pp[p] = f12d3p(fdwrap(callM, mvals, p), h = self.h, f0 = M)
                    # The [0] indicates that we are fitting the RMS force and not the RMSD
                    # (without the covariance, subtracting a mean force doesn't make sense.)
                    if all(M_p[p] == 0): continue
                    M0_M_p[p][0]  += P * M_p[p][0]
                    M0_Q_p[p][0]  += R * M_p[p][0]
                    #M0_M_pp[p][0] += P * M_pp[p][0]
                    #M0_Q_pp[p][0] += R * M_pp[p][0]
                    Xi_p        = 2 * X * M_p[p]
                    Xi_p[0]    *= boost
                    SPiXi_p[p] += P * Xi_p
                    SRiXi_p[p] += R * Xi_p
                    if not AHess: continue
                    if self.all_at_once:
                        M_pp[p] = ddM_all[i, p]
                    # This formula is more correct, but perhapsively convergence is slower.
                    #Xi_pq       = 2 * (M_p[p] * M_p[p] + X * M_pp[p])
                    # Gauss-Newton formula for approximate Hessian
                    Xi_pq       = 2 * (M_p[p] * M_p[p])
                    Xi_pq[0]   *= boost
                    SPiXi_pq[p,p] += P * Xi_pq
                    SRiXi_pq[p,p] += R * Xi_pq
                    for q in range(p):
                        if all(M_p[q] == 0): continue
                        if q not in self.pgrad: continue
                        Xi_pq          = 2 * M_p[p] * M_p[q]
                        Xi_pq[0]      *= boost
                        SPiXi_pq[p,q] += P * Xi_pq
                        SRiXi_pq[p,q] += R * Xi_pq

        # Dump energies and forces to disk.
        M_all_print = M_all.copy()
        if self.savg:
            M_all_print[:,0] -= np.average(M_all_print[:,0], weights=self.boltz_wts)
        if self.force:
            Q_all_print = np.hstack((col(self.eqm),self.fref))
        else:
            Q_all_print = col(self.eqm)
        if self.savg:
            QEtmp = np.array(Q_all_print[:,0]).flatten()
            Q_all_print[:,0] -= np.average(QEtmp, weights=self.boltz_wts)
        if self.attenuate: 
            QEtmp = np.array(Q_all_print[:,0]).flatten()
            Q_all_print[:,0] -= np.min(QEtmp)
            MEtmp = np.array(M_all_print[:,0]).flatten()
            M_all_print[:,0] -= np.min(MEtmp)
        if self.writelevel > 1:
            np.savetxt('M.txt',M_all_print)
            np.savetxt('Q.txt',Q_all_print)
        if self.writelevel > 0:
            EnergyComparison = np.hstack((col(Q_all_print[:,0]),col(M_all_print[:,0])))
            np.savetxt('QM-vs-MM-energies.txt',EnergyComparison)
            WeightComparison = np.hstack((col(Q_all_print[:,0]),col(self.boltz_wts)))
            np.savetxt('QM-vs-Wts.txt',WeightComparison)
        if self.force and self.writelevel > 1:
            # Write .xyz files which can be viewed in vmd.
            QMTraj = self.mol[:].atom_select(self.fitatoms)
            Mforce_obj = QMTraj[:]
            Qforce_obj = QMTraj[:]
            Mforce_print = np.array(M_all_print[:,1:3*nat+1])
            Qforce_print = np.array(Q_all_print[:,1:3*nat+1])
            Dforce_norm = np.array([np.linalg.norm(Mforce_print[i,:] - Qforce_print[i,:]) for i in range(NS)])
            MaxComp = np.max(np.abs(np.vstack((Mforce_print,Qforce_print)).flatten()))
            Mforce_print /= MaxComp
            Qforce_print /= MaxComp
            for i in range(NS):
                Mforce_obj.xyzs[i] = Mforce_print[i, :].reshape(-1,3)
                Qforce_obj.xyzs[i] = Qforce_print[i, :].reshape(-1,3)
            # if nat < len(self.qmatoms):
            #     Fpad = np.zeros((len(self.qmatoms) - nat, 3))
            #     Mforce_obj.xyzs[i] = np.vstack((Mforce_obj.xyzs[i], Fpad))
            #     Qforce_obj.xyzs[i] = np.vstack((Qforce_obj.xyzs[i], Fpad))
            if Mforce_obj.na != Mforce_obj.xyzs[0].shape[0]:
                print(Mforce_obj.na)
                print(Mforce_obj.xyzs[0].shape[0])
                warn_once('\x1b[91mThe printing of forces is not set up correctly.  Not printing forces.  Please report this issue.\x1b[0m')
            else:
                if self.writelevel > 1:
                    QMTraj.write('coords.xyz')
                    Mforce_obj.elem = ['H' for i in range(Mforce_obj.na)]
                    Mforce_obj.write('MMforce.xyz')
                    Qforce_obj.elem = ['H' for i in range(Qforce_obj.na)]
                    Qforce_obj.write('QMforce.xyz')
                    np.savetxt('Dforce_norm.dat', Dforce_norm)


        #==============================================================#
        #    STEP 3: Build the (co)variance matrix and invert it.      #
        # In the case of no covariance, this is just a diagonal matrix #
        # with the RMSD energy in [0,0] and the RMS gradient in [n, n] #
        #==============================================================#
        logger.debug("Done with snapshots, building objective function now\r")
        if (self.w_energy > 0.0 or self.w_force > 0.0 or self.w_netforce > 0.0 or self.w_torque > 0.0):
            wtot    = self.w_energy + self.w_force + self.w_netforce + self.w_torque
            EWt     = self.w_energy / wtot
            FWt     = self.w_force / wtot
            NWt     = self.w_netforce / wtot
            TWt     = self.w_torque / wtot
        else:
            EWt = 0.0
            FWt = 0.0
            NWt = 0.0
            TWt = 0.0
        # Build the weight vector/matrix, so the force contribution is suppressed by 1/3N
        if cv:
            WM      = np.zeros((NCP1,NCP1))
            WM[0,0] = np.sqrt(EWt)
            start   = 1
            block   = 3*nat
            end     = start + block
            for i in range(start, end):
                WM[i, i] = np.sqrt(FWt / block)
            if self.use_nft:
                start   = end
                block   = 3*nnf
                end     = start + block
                for i in range(start, end):
                    WM[i, i] = np.sqrt(NWt / block)
                start   = end
                block   = 3*ntq
                end     = start + block
                for i in range(start, end):
                    WM[i, i] = np.sqrt(TWt / block)
        else:
            WM      = np.zeros(NCP1)
            WM[0] = np.sqrt(EWt)
            if self.force:
                start   = 1
                block   = 3*nat
                end     = start + block
                WM[start:end] = np.sqrt(FWt / block)
                if self.use_nft:
                    start   = end
                    block   = 3*nnf
                    end     = start + block
                    WM[start:end] = np.sqrt(NWt / block)
                    start   = end
                    block   = 3*ntq
                    end     = start + block
                    WM[start:end] = np.sqrt(TWt / block)
        if cv:
            #==============================================================#
            #                      The covariance matrix                   #
            #                                                              #
            # Note: There is a detail here that might appear peculiar :    #
            # I'm building only one covariance matrix from both ensembles. #
            # This is the proper thing to do, believe me.                  #
            # The functional form looks like (X2_M + X2_Q)(C)^-1           #
            # I will build X2_M C^-1 and X2_Q C^-1 separately.             #
            #==============================================================#
            QBP  = self.qmboltz
            MBP  = 1 - self.qmboltz
            C    = MBP*(QQ_M-np.outer(Q0_M,Q0_M)/Z)/Z + QBP*(QQ_Q-np.outer(Q0_Q,Q0_Q)/Y)/Y
            Ci   = invert_svd(C)
            # Get rid of energy-force covariance
            for i in range(1,NCP1):
                Ci[0,i] = 0.0;
                Ci[i,0] = 0.0;
            WCiW = np.array(np.matrix(WM) * np.matrix(Ci) * np.matrix(WM)) # Weighted covariance matrix.
        else:
            # Here we're just using the variance.
            QBP  = self.qmboltz
            MBP  = 1 - self.qmboltz
            C    = MBP*(QQ_M-Q0_M*Q0_M/Z)/Z + QBP*(QQ_Q-Q0_Q*Q0_Q/Y)/Y
            if self.force_average:
                # Normalize over all atoms
                C[1:len(C)] = np.mean(C[1:len(C)])
            else:
                # Normalize over individual atoms (default)
                for i in range(1, len(C), 3):
                    C[i:i+3] = np.mean(C[i:i+3])
            Ci    = 1. / C
            WCiW = WM * Ci * WM
        #==============================================================#
        # STEP 4: Build the objective function and its derivatives.    #
        # Note that building derivatives with covariance is not yet    #
        # implemented, but can use external warpper to differentiate   #
        #==============================================================#
        if cv:
            X2_M  = build_objective(SPiXi,WCiW,Z,Q0_M,M0_M,NCP1)
            X2_Q  = build_objective(SRiXi,WCiW,Y,Q0_Q,M0_Q,NCP1)
        else:
            X2_M  = weighted_variance(SPiXi,WCiW,Z,X0_M,X0_M,NCP1,subtract_mean = self.savg)
            X2_Q  = weighted_variance(SRiXi,WCiW,Y,X0_Q,X0_Q,NCP1,subtract_mean = self.savg)
            # Print out all energy / force contributions, useful for debugging.
            # for i in range(XiAll.shape[0]):
            #     efctr = weighted_variance(XiAll[i],WCiW,Z,X0_M,X0_M,NCP1,subtract_mean = self.savg)
            #     WCiW1 = WCiW.copy()
            #     for j in range(1, len(WCiW1)):
            #         WCiW1[j] = 0.0
            #     ectr = weighted_variance(XiAll[i],WCiW1,Z,X0_M,X0_M,NCP1,subtract_mean = self.savg)
            #     print i, "ectr = %.3f efctr = %.3f" % (ectr, efctr)
            for p in self.pgrad:
                if not AGrad: continue
                X2_M_p[p] = weighted_variance(SPiXi_p[p],WCiW,Z,2*X0_M,M0_M_p[p],NCP1,subtract_mean = self.savg)
                X2_Q_p[p] = weighted_variance(SRiXi_p[p],WCiW,Y,2*X0_Q,M0_Q_p[p],NCP1,subtract_mean = self.savg)
                if not AHess: continue
                X2_M_pq[p,p] = weighted_variance2(SPiXi_pq[p,p],WCiW,Z,2*M0_M_p[p],M0_M_p[p],2*X0_M,M0_M_pp[p],NCP1,subtract_mean = self.savg)
                X2_Q_pq[p,p] = weighted_variance2(SRiXi_pq[p,p],WCiW,Y,2*M0_Q_p[p],M0_Q_p[p],2*X0_Q,M0_Q_pp[p],NCP1,subtract_mean = self.savg)
                for q in range(p):
                    if q not in self.pgrad: continue
                    X2_M_pq[p,q] = weighted_variance(SPiXi_pq[p,q],WCiW,Z,2*M0_M_p[p],M0_M_p[q],NCP1,subtract_mean = self.savg)
                    X2_Q_pq[p,q] = weighted_variance(SRiXi_pq[p,q],WCiW,Y,2*M0_Q_p[p],M0_Q_p[q],NCP1,subtract_mean = self.savg)
                    # Get the other half of the Hessian matrix.
                    X2_M_pq[q,p] = X2_M_pq[p,q]
                    X2_Q_pq[q,p] = X2_Q_pq[p,q]
        #==============================================================#
        #            STEP 5: Build the return quantities.              #
        #==============================================================#
        # The objective function
        X2   = MBP * X2_M    + QBP * X2_Q
        if not cv:
            # Derivatives of the objective function
            G = np.zeros(NP)
            H = np.zeros((NP,NP))
            for p in self.pgrad:
                if not AGrad: continue
                G[p] = MBP * X2_M_p[p] + QBP * X2_Q_p[p]
                if not AHess: continue
                for q in self.pgrad:
                    H[p,q] = MBP * X2_M_pq[p,q] + QBP * X2_Q_pq[p,q]
        # Energy error in kJ/mol
        if self.savg:
            E0_M = (2*Q0_M[0]*M0_M[0] - Q0_M[0]*Q0_M[0] - M0_M[0]*M0_M[0])/Z/Z;
            E0_Q = (2*Q0_Q[0]*M0_Q[0] - Q0_Q[0]*Q0_Q[0] - M0_Q[0]*M0_Q[0])/Y/Y;
        else:
            E0_M = 0.0
            E0_Q = 0.0

        if cv:
            dE     = MBP * np.sqrt(SPiXi[0,0]/Z + E0_M) + QBP * np.sqrt(SRiXi[0,0]/Y + E0_Q)
        else:
            dE     = MBP * np.sqrt(SPiXi[0]/Z + E0_M) + QBP * np.sqrt(SRiXi[0]/Y + E0_Q)

        if self.writelevel > 0:
            dE_print = (col(M_all_print[:,0]) - col(Q_all_print[:,0])) - (E0_M - E0_Q)
            ErrsvsWts = np.hstack((dE_print, col(self.boltz_wts)))
            np.savetxt('Errs-vs-Wts.txt',ErrsvsWts)

        # Fractional energy error.
        dEfrac = MBP * np.sqrt((SPiXi[0]/Z + E0_M) / (QQ_M[0]/Z - Q0_M[0]**2/Z/Z)) + QBP * np.sqrt((SRiXi[0]/Y + E0_Q) / (QQ_Q[0]/Y - Q0_Q[0]**2/Y/Y))
        # Absolute and Fractional force error.
        if self.force:
            dF = MBP * dF_M / Z + QBP * dF_Q / Y
            qF = MBP * qF_M / Z + QBP * qF_Q / Y
            dFfrac = MBP * (dF_M/qF_M) + QBP * (dF_Q/qF_Q)
        if self.use_nft:
            dN = MBP * dN_M / Z + QBP * dN_Q / Y
            qN = MBP * qN_M / Z + QBP * qN_Q / Y
            dNfrac = MBP * dN_M / qN_M + QBP * dN_Q / qN_Q
            dT = MBP * dT_M / Z + QBP * dT_Q / Y
            qT = MBP * qT_M / Z + QBP * qT_Q / Y
            dTfrac = MBP * dT_M / qT_M + QBP * dT_Q / qT_Q
        # Save values to qualitative indicator if not inside finite difference code.
        if not in_fd():
            # Contribution from energy and force parts.
            self.e_ctr = (MBP * weighted_variance(np.array([SPiXi[0]]),np.array([WCiW[0]]),Z,X0_M,X0_M,NCP1,subtract_mean = self.savg) + 
                          QBP * weighted_variance(np.array([SRiXi[0]]),np.array([WCiW[0]]),Y,X0_Q,X0_Q,NCP1,subtract_mean = self.savg))
            self.e_ref = MBP * np.sqrt(QQ_M[0]/Z - Q0_M[0]**2/Z/Z) + QBP * np.sqrt((QQ_Q[0]/Y - Q0_Q[0]**2/Y/Y))
            self.e_err = dE
            self.e_err_pct = dEfrac
            if self.force:
                self.f_ctr = (MBP * weighted_variance(SPiXi[1:1+3*nat],WCiW[1:1+3*nat],Z,X0_M,X0_M,NCP1,subtract_mean = False) + 
                              QBP * weighted_variance(SRiXi[1:1+3*nat],WCiW[1:1+3*nat],Y,X0_Q,X0_Q,NCP1,subtract_mean = False))
                self.f_ref = qF
                self.f_err = dF
                self.f_err_pct = dFfrac
            if self.use_nft:
                self.nf_ctr = (MBP * weighted_variance(SPiXi[1+3*nat:1+3*nat+3*nnf],WCiW[1+3*nat:1+3*nat+3*nnf],Z,X0_M,X0_M,NCP1,subtract_mean = False) + 
                               QBP * weighted_variance(SRiXi[1+3*nat:1+3*nat+3*nnf],WCiW[1+3*nat:1+3*nat+3*nnf],Y,X0_Q,X0_Q,NCP1,subtract_mean = False))
                self.nf_ref = qN
                self.nf_err = dN
                self.nf_err_pct = dNfrac
                self.tq_ctr = (MBP * weighted_variance(SPiXi[1+3*nat+3*nnf:1+3*nat+3*nnf+3*ntq],WCiW[1+3*nat+3*nnf:1+3*nat+3*nnf+3*ntq],Z,X0_M,X0_M,NCP1,subtract_mean = False) + 
                               QBP * weighted_variance(SRiXi[1+3*nat+3*nnf:1+3*nat+3*nnf+3*ntq],WCiW[1+3*nat+3*nnf:1+3*nat+3*nnf+3*ntq],Y,X0_Q,X0_Q,NCP1,subtract_mean = False))
                self.tq_ref = qT
                self.tq_err = dT
                self.tq_err_pct = dTfrac
            pvals = self.FF.make(mvals) # Write a force field that isn't perturbed by finite differences.
        if cv:
            Answer = {'X':X2, 'G':np.zeros(self.FF.np), 'H':np.zeros((self.FF.np,self.FF.np))}
        else:
            Answer = {'X':X2, 'G':G, 'H':H}
        return Answer

    def get_resp(self, mvals, AGrad=False, AHess=False):
        """ Electrostatic potential fitting.  Implements the RESP objective function.  (In Python so obviously not optimized.) """
        if (self.w_resp == 0.0):
            AGrad = False
            AHess = False
        Answer = {}
        pvals = self.FF.make(mvals)

        # Build the distance matrix for ESP fitting.
        self.invdists = self.build_invdist(mvals)

        NS = self.ns
        NP = self.FF.np
        Z = 0
        Y = 0
        def new_charges(mvals_):
            """ Return the charges acting on the system. """
            logger.debug("\r")
            pvals = self.FF.make(mvals_)
            return self.engine.get_charges()
            
            # Need to update the positions of atoms, if there are virtual sites.
            # qvals = [pvals[i] for i in self.FF.qmap]
            # # All of a sudden I need the number of virtual sites.
            # qatoms = np.zeros(self.nparticles)
            # for i, jj in enumerate(self.FF.qid):
            #     for j in jj:
            #         qatoms[j] = qvals[i]
            # return qatoms

        # Obtain a derivative matrix the stupid way
        charge0 = new_charges(mvals)
        if AGrad:
            # dqPdqM = []
            # for i in range(NP):
            #     print "Now working on parameter number", i
            #     dqPdqM.append(f12d3p(fdwrap(new_charges,mvals,i), h = self.h)[0])
            # dqPdqM = np.matrix(dqPdqM).T
            dqPdqM = np.matrix([(f12d3p(fdwrap(new_charges,mvals,i), h = self.h, f0 = charge0)[0] if i in self.pgrad else np.zeros_like(charge0)) for i in range(NP)]).T
        xyzs = np.array(self.mol.xyzs)
        espqvals = np.array(self.espval)
        espxyz   = np.array(self.espxyz)

        ddVdqPdVS = {}
        # Second derivative of the inverse distance matrix with respect to the virtual site position
        dddVdqPdVS2 = {}
        if AGrad:
            for p in self.pgrad:
                if 'VSITE' in self.FF.plist[p]:
                    ddVdqPdVS[p], dddVdqPdVS2[p] = f12d3p(fdwrap(self.build_invdist,mvals,p), h = self.h, f0 = self.invdists)
        X = 0
        Q = 0
        D = 0
        G = np.zeros(NP)
        H = np.zeros((NP, NP))
        for i in range(self.ns):
            P   = self.boltz_wts[i]
            Z  += P
            dVdqP   = np.matrix(self.invdists[i])
            espqval = espqvals[i]
            espmval = dVdqP * col(new_charges(mvals))
            desp    = flat(espmval) - espqval
            X      += P * np.dot(desp, desp) / self.nesp
            Q      += P * np.dot(espqval, espqval) / self.nesp
            D      += P * (np.dot(espqval, espqval) / self.nesp - (np.sum(espqval) / self.nesp)**2)
            if AGrad:
                dVdqM   = (dVdqP * dqPdqM).T
                for p, vsd in ddVdqPdVS.items():
                    dVdqM[p,:] += flat(vsd[i] * col(new_charges(mvals)))
                G      += flat(P * 2 * dVdqM * col(desp)) / self.nesp
                if AHess:
                    d2VdqM2 = np.zeros(dVdqM.shape)
                    for p, vsd in dddVdqPdVS2.items():
                        d2VdqM2[p,:] += flat(vsd[i] * col(new_charges(mvals)))
                    H      += np.array(P * 2 * (dVdqM * dVdqM.T + d2VdqM2 * col(desp))) / self.nesp
        # Redundant but we keep it anyway
        D /= Z
        X /= Z
        X /= D
        G /= Z
        G /= D
        H /= Z
        H /= D
        Q /= Z
        Q /= D
        if not in_fd():
            self.esp_err = np.sqrt(X)
            self.esp_ref = np.sqrt(Q)
            self.esp_err_pct = self.esp_err / self.esp_ref
            
        # Following is the restraint part
        # RESP hyperbola "strength" parameter; 0.0005 is weak, 0.001 is strong
        # RESP hyperbola "tightness" parameter; don't need to change this
        a = self.resp_a
        b = self.resp_b
        q = new_charges(mvals)
        R   = a*np.sum((q**2 + b**2)**0.5 - b)
        dR  = a*q*(q**2 + b**2)**-0.5
        ddR = a*b**2*(q**2 + b**2)**-1.5
        self.respterm = R
        X += R
        if AGrad:
            G += flat(dqPdqM.T * col(dR))
            if AHess:
                H += np.diag(flat(dqPdqM.T * col(ddR)))

        if not in_fd():
            self.esp_ctr = X
            
        Answer = {'X':X,'G':G,'H':H}
        return Answer

    def get(self, mvals, AGrad=False, AHess=False):
        Answer = {'X':0.0, 'G':np.zeros(self.FF.np), 'H':np.zeros((self.FF.np, self.FF.np))}
        tw = self.w_energy + self.w_force + self.w_netforce + self.w_torque + self.w_resp
        if tw > 0.0:
            w_ef = (self.w_energy + self.w_force + self.w_netforce + self.w_torque) / tw
            w_resp = self.w_resp / tw
        else:
            w_ef = 0.0
            w_resp = 0.0
        if self.energy or self.force:
            Answer_EF = self.get_energy_force(mvals, AGrad, AHess)
            for i in Answer_EF:
                Answer[i] += w_ef * Answer_EF[i]
        if self.resp:
            Answer_ESP = self.get_resp(mvals, AGrad, AHess)
            for i in Answer_ESP:
                Answer[i] += w_resp * Answer_ESP[i]
        if not any([self.energy, self.force, self.resp]):
            logger.error("Ab initio fitting must have at least one of: Energy, Force, ESP\n")
            raise RuntimeError
        if not in_fd():
            self.objective = Answer['X']
        return Answer

def weighted_variance(SPiXi,WCiW,Z,L,R,NCP1,subtract_mean=True):
    """ A more generalized version of build_objective which is
    callable for derivatives, but the covariance is not there anymore. """
    # These are the functions that we are building.
    X2        = 0.0
    # Divide by Z to normalize
    XiZ       = SPiXi/Z
    # Subtract out the average energy.
    if subtract_mean:
        XiZ[0] -= (L[0] * R[0])/Z/Z
    # Return the answer.
    X2      = np.dot(XiZ.flatten(),WCiW.flatten())
    return X2

def weighted_variance2(SPiXi,WCiW,Z,L,R,L2,R2,NCP1,subtract_mean=True):
    """ A bit of a hack, since we have to subtract out two mean quantities to get Hessian elements. """
    # These are the functions that we are building.
    X2        = 0.0
    # Divide by Z to normalize
    XiZ       = SPiXi/Z
    # Subtract out the average energy.
    if subtract_mean:
        XiZ[0] -= (L[0] * R[0])/Z/Z
        XiZ[0] -= (L2[0] * R2[0])/Z/Z
    # Return the answer.
    X2      = np.dot(XiZ.flatten(),WCiW.flatten())
    return X2

def build_objective(SPiXi,WCiW,Z,Q0,M0,NCP1,subtract_mean=True):

    """ This function builds an objective function (number) from the
    complicated polytensor and covariance matrices. """

    # These are the functions that we are building.
    X2    = 0.0
    # Divide by Z to normalize
    XiZ       = SPiXi/Z
    if subtract_mean:
        # Subtract out the zero-point energy gap
        XiZ[0,0] -= (M0[0]*M0[0] + Q0[0]*Q0[0] - 2*Q0[0]*M0[0])/Z/Z
        for i in range(1,NCP1):
            XiZ[0,i] -= (M0[i]*M0[0] + Q0[i]*Q0[0] - 2*Q0[i]*M0[0])/Z/Z
            XiZ[i,0] -= (M0[0]*M0[i] + Q0[0]*Q0[i] - 2*Q0[0]*M0[i])/Z/Z
    ### This is the objective function! LAAAAA ###
    X2      = np.dot(XiZ.flatten(),WCiW.flatten())
    return X2




#!/usr/bin/env python

# Very simple file format conversion script, works in 90% of situations
# and good for saving keystrokes.

from forcebalance.molecule import Molecule
from sys import argv

def main():
    topfnm = argv[3] if len(argv) >= 4 else None
    M = Molecule(argv[1], top=topfnm)
    M.write(argv[2])

if __name__ == "__main__":
    main()



""" @package forcebalance.qchemio Q-Chem input file parser. """

from builtins import str
from builtins import range
import os
from re import match, sub
from forcebalance import BaseReader
from forcebalance.nifty import *
from forcebalance.molecule import Molecule

from forcebalance.output import getLogger
logger=getLogger(__name__)

## Types of counterpoise correction
#cptypes = [None, 'BASS', 'BASSP']
## Types of NDDO correction
ndtypes = [None]

##Section -> Interaction type dictionary.
#fdict = {
#    'basis'  : bastypes    }

##Interaction type -> Parameter Dictionary.
pdict = {'BASS':{0:'A', 1:'C'},
         'BASSP' :{0:'A', 1:'B', 2:'C'}
         }

class QCIn_Reader(BaseReader):
    """Finite state machine for parsing Q-Chem input files.
    
    """
    
    def __init__(self,fnm):
        # Initialize the superclass. :)
        super(QCIn_Reader,self).__init__(fnm)
        self.atom  = ''
        self.snum  = -1
        self.cnum  = -1
        self.shell = None
        self.pdict = pdict
        
    def feed(self, line):
        """ Feed in a line.

        @param[in] line     The line of data

        """
        line       = line.split('!')[0].strip()
        s          = line.split()
        #self.itype = None
        self.ln   += 1
        # No sense in doing anything for an empty line or a comment line.
        if len(s) == 0 or match('^!',line): return None, None
        # Now go through all the cases.
        if match('^\$',line):
            # Makes a word like "atoms", "bonds" etc.
            self.sec = sub('^\$','',line)
        elif self.sec == 'basis':
            if match('^[A-Za-z][a-z]* +0$',line):
                self.atom = s[0]
                self.snum = -1
            elif match('^[SPDFGH]P? +[0-9]+ +1\.0+$',line):
                self.snum += 1
                self.cnum  = -1
                self.shell = s[0]
                self.itype = 'BAS'+self.shell
                logger.info(line + '\n')
            elif all([isfloat(i) for i in s]):
                self.cnum += 1
                logger.info("%s %s %s\n" % (line, str(self.snum), str(self.cnum)))
                self.suffix = '_at%s.sh%i.cf%i' % (self.atom,self.snum,self.cnum)

def QChem_Dielectric_Energy(fnm,wq):
    QCIn = Molecule(fnm)
    for i in range(QCIn.na):
        # Q-Chem crashes if it doesn't recognize the chemical element
        if QCIn.Data['elem'][i] in ['M','L']:
            QCIn.Data['elem'][i] = 'He'
    CalcDir=os.path.splitext(fnm)[0]+".d"
    GoInto(CalcDir)
    digits = len(str(QCIn.ns))
    for i in range(QCIn.ns):
        sdir = "%%0%ii" % digits % i
        GoInto(sdir)
        QCIn.write("qchem.in",selection=i)
        queue_up(wq,"qchem40 qchem.in qchem.out",input_files=["qchem.in"],output_files=["qchem.out"],verbose=False)
        Leave(sdir)
    wq_wait(wq,verbose=False)
    PCM_Energies = []
    for i in range(QCIn.ns):
        sdir = "%%0%ii" % digits % i
        GoInto(sdir)
        for line in open("qchem.out"):
            if "PCM electrostatic energy" in line:
                PCM_Energies.append(float(line.split()[-2]))
        Leave(sdir)
    Leave(CalcDir)
    return np.array(PCM_Energies) * 2625.5



from builtins import str
from builtins import object
import forcebalance
from uuid import uuid1 as uuid
import os

"""
This file contains classes that interface with forcebalance and act as an intermediary
between the GUI frontend and lower level calculation elements.

Objects that need to be displayed should implement a display() function that returns
a type that can be recognized by the widget. Generic display function that returns a
string filled with an object's properties is provided by ForceBalanceObject, but
more complex data structures like dictionaries provide information on structure that
can be used by the widget in creating more complicated views.
"""

class ForceBalanceObject(object):
    """Generic forcebalance object with some identifying information"""
    def __init__(self, name='', type='object'):
        self.properties = dict()
        self.properties['active'] = False
        self.properties['name']=name
        self.properties['type'] = type
        self.properties['id'] = str(uuid())

    def __getitem__(self, item):
        return self.properties[item]

    def __setitem__(self, key, value):
        self.properties[key]=value

    def display(self, verbose = False):
        s = ''
        for key in self.properties.keys():
            s+= "%s : %s\n" % (key, str(self.properties[key]))
        return s

class CalculationObject(ForceBalanceObject):
    """Represents a set of forcebalance objects that together make up a calculation"""
    def __init__(self, filename=None):
        super(CalculationObject,self).__init__(type='calculation')
        self.properties['options']=[]
        self.properties['targets']=[]
        self.properties['forcefield']=None
        self.properties['result']=None

        self.properties['_expand']=True
        self.properties['_expand_targets']=True

        cwd=os.getcwd()
        os.chdir(os.path.dirname(filename))

        self.opts, self.tgt_opts = forcebalance.parser.parse_inputs(filename)
        self.properties['options'] = OptionObject(self.opts, os.path.basename(filename))

        self.properties['name'] = self.properties['options']['name']

        for target in self.tgt_opts:
            self.properties['targets'].append(TargetObject(target))

        if filename:
            self.properties['forcefield'] = ForcefieldObject(self.opts)

        os.chdir(cwd)

    def addTarget(self, opts = forcebalance.parser.tgt_opts_defaults.copy()):
        self.properties['targets'].append(TargetObject(opts))

    def writeOptions(self, filename):
        with open('~' + filename,'w') as f:
            f.write("$options\n")
            for option in self.properties['options'].opts.keys():
                f.write("%s %s\n" % (option, self.properties['options'].opts[option]))
            f.write("$end\n")
            
            for target in self.properties['targets']:
                f.write("\n$target\n")
                for option in target.opts:
                    f.write("%s %s\n" % (option, target.opts[option]))
                f.write("$end\n")

    def display(self, verbose = False):
        properties = {
            'forcefield' : self.properties['forcefield'],
            'input file' : self.properties['options']['name']}

        for n, target in enumerate(self.properties['targets']):
            properties.update({"target%d" % n : target['name']})

        return properties

    def run(self):
        try:
            #options, tgt_opts = parse_inputs(input_file)
            ## The force field component of the project
            #forcefield  = FF(options)
            ## The objective function
            objective   = forcebalance.objective.Objective(self.opts, self.tgt_opts, self.properties['forcefield'].forcefield)
            ## The optimizer component of the project
            optimizer   = forcebalance.optimizer.Optimizer(self.opts, objective, self.properties['forcefield'].forcefield)
            ## Actually run the optimizer.
            optimizer.Run()
            
            
            resultopts = self.opts.copy()
            resultopts.update({"ffdir" : "result"})
            
            # temporarily silence nifty and forcefield while reading the results forcefield
            forcebalance.output.getLogger("forcebalance.forcefield").disabled = True
            forcebalance.output.getLogger("forcebalance.nifty").propagate = True
            self.properties['result'] = ForcefieldObject(resultopts)
            forcebalance.output.getLogger("forcebalance.forcefield").disabled = False
            forcebalance.output.getLogger("forcebalance.nifty").disabled = False
        except:
            import traceback
            logger = forcebalance.output.getLogger("forcebalance")
            logger.critical("Calculation failed\n%s" % traceback.format_exc())
        
# maybe the current implementation of TargetObject should be merged here
# to keep all options in the same place?
class OptionObject(ForceBalanceObject):
    """Represents general options parsed from an input file"""
    def __init__(self, opts=None, name="unknown options file"):
        super(OptionObject,self).__init__(type = 'options')

        if not opts: self.opts = forcebalance.parser.gen_opts_defaults
        else: self.opts = opts

        self.properties['name'] = name

    def display(self, verbose = False):
        options = dict()
        default_options = dict()
        for key in self.opts.keys():
            if not self.isDefault(key):
                options[key]=self.opts[key]
            else: default_options[key]=self.opts[key]
        return (options, default_options)

    def resetOption(self, option):
        try:
            self.opts[option] = forcebalance.parser.gen_opts_defaults[option]
        except:
            self.opts[option] = None        

    def setOption(self, option, value):
        self.opts[option] = value

    def getOptionHelp(self, option):
        message = "No help available for option '%s'" % option

        for opt_type in forcebalance.parser.gen_opts_types:
            if option in forcebalance.parser.gen_opts_types[opt_type].keys():
                message = option + ": "
                message += forcebalance.parser.gen_opts_types[opt_type][option][2]
                message +="\nApplicable to: "
                message += forcebalance.parser.gen_opts_types[opt_type][option][3]
        return message

    def isDefault(self, option):
        return option in forcebalance.parser.gen_opts_defaults.keys() and self.opts[option] == forcebalance.parser.gen_opts_defaults[option]

class TargetObject(ForceBalanceObject):
    """Currently holds target option information. In the future, might also hold actual target objects"""
    def __init__(self, tgt_opt):
        super(TargetObject,self).__init__(type='target')
        self.opts = tgt_opt
        self.properties.update(self.opts)

    def display(self, verbose=0):
        options = dict()
        default_options = dict()
        for key in self.opts.keys():
            if not self.isDefault(key):
                options[key]=self.opts[key]
            else: default_options[key]=self.opts[key]
        return (options, default_options)

    def resetOption(self, option):
        try:
            self.opts[option] = forcebalance.parser.tgt_opts_defaults[option]
        except:
            self.opts[option] = None

    def setOption(self, option, value):
        self.opts[option] = value

    def getOptionHelp(self, option):
        message = "No help available for option '%s'" % option

        for opt_type in forcebalance.parser.tgt_opts_types:
            if option in forcebalance.parser.tgt_opts_types[opt_type].keys():
                message = option + ": "
                message += forcebalance.parser.tgt_opts_types[opt_type][option][2]
                message +="\nApplicable to: "
                message += forcebalance.parser.tgt_opts_types[opt_type][option][3]
        return message

    def isDefault(self, option):
        return option in forcebalance.parser.tgt_opts_defaults.keys() and self.opts[option] == forcebalance.parser.tgt_opts_defaults[option]

class ForcefieldObject(ForceBalanceObject):
    """Represents a forcebalance forcefield object"""
    def __init__(self, opts):
        super(ForcefieldObject, self).__init__(type = 'forcefield')
        self.opts = opts
        self.forcefield = forcebalance.forcefield.FF(self.opts)
        self.properties["name"] = self.forcefield.fnms[0]

    def display(self, verbose = False):
        s=''
        s+= "name : " + self.properties["name"] + '\n'
        s+= "files : " + str(self.forcefield.fnms) + '\n'
        s+= "number of parameters : " + str(self.forcefield.np) + '\n'
        s+= "map : \n"
        for value in self.forcefield.map:
            s+= '\t' + str(value) + '\n'
        s+= "pvals : " + str(self.forcefield.pvals0) + '\n'
        return s



""" @package forcebalance.custom_io Custom force field parser.

We take advantage of the sections in GROMACS and the 'interaction
type' concept, but these interactions are not supported in GROMACS;
rather, they are computed within our program.

@author Lee-Ping Wang
@date 12/2011
"""

from re import match, sub
from forcebalance import BaseReader

## Types of counterpoise correction
cptypes = [None, 'CPGAUSS', 'CPEXPG', 'CPGEXP']
## Types of NDDO correction
ndtypes = [None]

##Section -> Interaction type dictionary.
fdict = {
    'counterpoise'  : cptypes    }

##Interaction type -> Parameter Dictionary.
pdict = {'CPGAUSS':{3:'A', 4:'B', 5:'C'},
         'CPGEXP' :{3:'A', 4:'B', 5:'G', 6:'X'},
         'CPEXPG' :{3:'A1', 4:'B', 5:'X0', 6:'A2'}
         }

class Gen_Reader(BaseReader):
    """Finite state machine for parsing custom GROMACS force field files.

    This class is instantiated when we begin to read in a file.  The
    feed(line) method updates the state of the machine, giving it
    information like the residue we're currently on, the nonbonded
    interaction type, and the section that we're in.  Using this
    information we can look up the interaction type and parameter type
    for building the parameter ID.
    
    """
    
    def __init__(self,fnm):
        # Initialize the superclass. :)
        super(Gen_Reader,self).__init__(fnm)
        ## The current section that we're in
        self.sec = None
        ## The parameter dictionary (defined in this file)
        self.pdict  = pdict

    def feed(self, line):
        """ Feed in a line.

        @param[in] line     The line of data

        """
        s          = line.split()
        atom       = []
        self.itype = None
        self.ln   += 1
        # No sense in doing anything for an empty line or a comment line.
        if len(s) == 0 or match('^;',line): return None, None
        # Now go through all the cases.
        if match('^\[.*\]',line):
            # Makes a word like "atoms", "bonds" etc.
            self.sec = sub('[\[\] \n]','',line)
        elif self.sec == 'counterpoise':
            self.itype = cptypes[int(s[2])]
            atom = [s[0],s[1]]
        elif self.sec == 'NDDO':
            # NDDO hasn't been tested since the refactoring.
            self.itype = '_'.join(['NDDO', s[0], s[1]])
        else:
            return [],"Confused"
        if len(atom) > 1 and atom[0] > atom[-1]:
            # Enforce a canonical ordering of the atom labels in a parameter ID
            atom = atom[::-1]
        self.suffix = ''.join(atom)



#!/usr/bin/env python

from __future__ import print_function
from builtins import zip
import numpy as np
import sys
import os
from re import match

#==========================================================#
# GAFF Explanation Script, Lee-Ping Wang           05/2012 #
#                                                          #
# The purpose of this script is to explain the atom types  #
# that come from an Antechamber run.  Sometimes the atom   #
# types are 'wrong' and we have to correct them.           #
#                                                          #
# The script works by reading in the antechamber output    #
# mol2 file (it doesn't work on other file types yet.)     #
# It parses the mol2 file for the coordinates / atomtypes  #
# and cross-references them with the gaff.dat data file.   #
#                                                          #
# Then it abbreviates the descriptions and prints them     #
# to a VMD script file along with the atomic coordinates.  #
# When this file is executed in VMD using 'vmd -e .runvmd' #
# it will load the molecule and label all of the atomtypes #
# using the abbrevaited description.                       #
#==========================================================#

try:
    DataFile=os.path.join(os.environ['AMBERHOME'],"dat/leap/parm/gaff.dat")
except:
    raise Exception("To make sure we find the gaff.dat file, please make sure the $AMBERHOME environment variable is set.")
if not os.path.exists(DataFile):
    raise Exception("gaff.dat not found!")
# Lee-Ping's gaff.dat
# DataFile="/home/leeping/opt/amber/dat/leap/parm/gaff.dat"
Atypes = {}

def isfloat(word):
    """Matches ANY number; it can be a decimal, scientific notation, integer, or what have you"""
    return match('^[-+]?[0-9]*\.?[0-9]*([eEdD][-+]?[0-9]+)?$',word)

for line in open(DataFile).readlines():
    sline = line.split()
    if len(sline) == 0:
        break
    if isfloat(sline[1]) and isfloat(sline[2]):
        # Following is a bunch of things to reduce the text description size.
        dline = ' '.join(sline[3:])
        dline = dline.replace("identical to","=")
        dline = dline.replace('in pure aromatic systems', 'pure arom')
        dline = dline.replace('in non-pure aromatic systems', 'nonpure arom')
        dline = dline.replace('in conjugated systems', 'conj')
        dline = dline.replace('in triangle systems', 'triangle')
        dline = dline.replace('in square systems', 'square')
        dline = dline.replace('Head Sp2 C that connect two rings in biphenyl sys.', 'bridgehead C')
        dline = dline.replace('H bonded to aliphatic carbon without electrwd. group', 'H-C (aliph) 0 EWG')
        dline = dline.replace('H bonded to aliphatic carbon with 1 electrwd. group', 'H-C (aliph) 1 EWG')
        dline = dline.replace('H bonded to aliphatic carbon with 2 electrwd. group', 'H-C (aliph) 2 EWG')
        dline = dline.replace('H bonded to aliphatic carbon with 3 electrwd. group', 'H-C (aliph) 3 EWG')
        dline = dline.replace('H bonded to C next to positively charged group', 'H-C (positive), 3 EWG')
        dline = dline.replace('H bonded to non-sp3 carbon with', 'H-C (non-sp3),')
        dline = dline.replace('H bonded to aromatic carbon', 'H-C (arom),')
        dline = dline.replace('H bonded to nitrogen atoms', 'H-N')
        dline = dline.replace('with two connected atoms', '2 bonds')
        dline = dline.replace('with three connected atoms', '3 bonds')
        dline = dline.replace('with four connected atoms', '4 bonds')
        dline = dline.replace('carbons', 'C')
        dline = dline.replace('carbon ', 'C ')
        Atypes[sline[0]] = '\ '.join(dline.split())

Descriptions = []
XYZs = []

if len(sys.argv) != 2:
    print("Usage: %s molecule.mol2" % __file__)
    print("Exiting...")
    sys.exit()

for line in os.popen("awk '/ATOM/,/BOND/' %s" % sys.argv[1]):
    sline = line.split()
    if len(sline) == 9:
        try:
            Descriptions.append("\ \ \ \[%s\]\ " % sline[5] + Atypes[sline[5]])
        except:
            raise KeyError("The atom type %s is not in gaff.dat" % sline[5])
        XYZs.append([float(i) for i in sline[2:5]])

#graphics top text {23.4840 -7.0510 10.1000} 'Sniffy'

vmdfnm = '.runvmd'
OutVMD = open(vmdfnm,'w')

header = """
#==================================================#
#           Global settings for display            #
#==================================================#
axes location Off
display rendermode GLSL
color Display Background white
display projection Orthographic
display depthcue off
display nearclip set 0.010000
material change opacity Ghost 0.000000

# Load the new molecule
mol new {xyzname} type mol2 waitfor all

mol modstyle 0 0 CPK 1.000000 0.300000 25.000000 25.000000
"""

print(header.format(xyzname=sys.argv[1]), file=OutVMD)

for xyz, desc in zip(XYZs, Descriptions):
    print("graphics top text {% .3f % .3f % .3f} %s size 0.4" % (xyz[0],xyz[1],xyz[2],desc), file=OutVMD)

OutVMD.close()

RunVMD = False
# Now Actually run VMD.
if RunVMD:
    os.system('vmd -e %s' % vmdfnm)
else:
    print("Finished analyzing mol2 file.  Now run vmd with the following command: vmd -e %s" % vmdfnm)



""" @package forcebalance.interaction Interaction energy fitting module.

@author Lee-Ping Wang
@date 05/2012
"""
from __future__ import division

from builtins import str
from builtins import range
import os
import shutil
import numpy as np
from forcebalance.nifty import col, eqcgmx, flat, floatornan, fqcgmx, invert_svd, kb, printcool, bohr2ang, commadash, uncommadash, printcool_dictionary
from forcebalance.target import Target
from forcebalance.molecule import Molecule, format_xyz_coord
from re import match, sub
import subprocess
from subprocess import PIPE
from forcebalance.finite_difference import fdwrap, f1d2p, f12d3p, in_fd
from collections import OrderedDict

from forcebalance.output import getLogger
logger = getLogger(__name__)

class Interaction(Target):

    """ Subclass of Target for fitting force fields to interaction energies.

    Currently TINKER is supported.

    We introduce the following concepts:
    - The number of snapshots
    - The reference interaction energies and the file they belong in (qdata.txt)

    This subclass contains the 'get' method for building the objective
    function from any simulation software (a driver to run the program and
    read output is still required)."""

    def __init__(self,options,tgt_opts,forcefield):
        # Initialize the SuperClass!
        super(Interaction,self).__init__(options,tgt_opts,forcefield)

        #======================================#
        # Options that are given by the parser #
        #======================================#

        ## Number of snapshots
        self.set_option(tgt_opts,'shots','ns')
        ## Do we call Q-Chem for dielectric energies? (Currently needs to be fixed)
        self.set_option(tgt_opts,'do_cosmo','do_cosmo')
        ## Do we put the reference energy into the denominator?
        self.set_option(tgt_opts,'cauchy','cauchy')
        ## Do we put the reference energy into the denominator?
        self.set_option(tgt_opts,'attenuate','attenuate')
        ## Divide by the number of snapshots?
        self.set_option(tgt_opts, 'normalize')
        ## What is the energy denominator?
        self.set_option(tgt_opts,'energy_denom','energy_denom')
        ## Set fragment 1
        self.set_option(tgt_opts,'fragment1','fragment1')
        if len(self.fragment1) == 0:
            logger.error('You need to define the first fragment using the fragment1 keyword\n')
            raise RuntimeError
        self.select1 = np.array(uncommadash(self.fragment1))
        ## Set fragment 2
        self.set_option(tgt_opts,'fragment2','fragment2')
        if len(self.fragment2) != 0:
            self.select2 = np.array(uncommadash(self.fragment2))
        else:
            self.select2 = None
        ## Set upper cutoff energy
        self.set_option(tgt_opts,'energy_upper','energy_upper')
        ## Option for how much data to write to disk.
        self.set_option(tgt_opts,'writelevel','writelevel')
        #======================================#
        #     Variables which are set here     #
        #======================================#
        ## LPW 2018-02-11: This is set to True if the target calculates
        ## a single-point property over several existing snapshots.
        self.loop_over_snapshots = True
        ## Reference (QM) interaction energies
        self.eqm           = []
        ## Snapshot label, useful for graphing
        self.label         = []
        ## The qdata.txt file that contains the QM energies and forces
        self.qfnm = os.path.join(self.tgtdir,"qdata.txt")
        self.e_err = 0.0
        self.e_err_pct = None
        ## Read in the trajectory file
        self.mol = Molecule(os.path.join(self.root,self.tgtdir,self.coords),
                            top=(os.path.join(self.root,self.tgtdir,self.pdb) if hasattr(self, 'pdb') else None), build_topology=False if self.coords.endswith('.pdb') else True)
        if self.ns != -1:
            self.mol = self.mol[:self.ns]
        self.ns = len(self.mol)
        if self.select2 is None:
            self.select2 = [i for i in range(self.mol.na) if i not in self.select1]
            logger.info('Fragment 2 is the complement of fragment 1 : %s\n' % (commadash(self.select2)))
        ## Build keyword dictionaries to pass to engine.
        engine_args = OrderedDict(list(self.OptionDict.items()) + list(options.items()))
        del engine_args['name']
        self.engine = self.engine_(target=self, mol=self.mol, **engine_args)
        ## Read in the reference data
        self.read_reference_data()
        logger.info("The energy denominator is: %s kcal/mol\n"  % str(self.energy_denom))
        denom = self.energy_denom
        # Create the denominator.
        if self.cauchy:
            self.divisor = np.sqrt(self.eqm**2 + denom**2)
            if self.attenuate:
                logger.error('attenuate and cauchy are mutually exclusive\n')
                raise RuntimeError
        elif self.attenuate:
            # Attenuate only large repulsions.
            self.divisor = np.zeros(len(self.eqm))
            for i in range(len(self.eqm)):
                if self.eqm[i] < denom:
                    self.divisor[i] = denom
                else:
                    self.divisor[i] = np.sqrt(denom**2 + (self.eqm[i]-denom)**2)
        else:
            self.divisor = np.ones(len(self.eqm)) * denom

        if self.cauchy:
            logger.info("Each contribution to the interaction energy objective function will be scaled by 1.0 / ( energy_denom**2 + reference**2 )\n")
        if self.energy_upper > 0:
            ecut = self.energy_upper
            self.prefactor = 1.0 * (self.eqm < ecut)
            logger.info("Interactions more repulsive than %s will not be fitted (%i/%i excluded) \n" % (str(self.energy_upper), sum(self.eqm > ecut), len(self.eqm)))
        else:
            self.prefactor = np.ones(len(self.eqm))
        if self.normalize:
            self.prefactor /= len(self.prefactor)

    def read_reference_data(self):

        """ Read the reference ab initio data from a file such as qdata.txt.

        After reading in the information from qdata.txt, it is converted
        into kcal/mol.

        """
        # Parse the qdata.txt file
        for line in open(os.path.join(self.root,self.qfnm)):
            sline = line.split()
            if len(sline) == 0: continue
            elif sline[0] == 'INTERACTION':
                self.eqm.append(float(sline[1]))
            elif sline[0] == 'LABEL':
                self.label.append(sline[1])
            if all(len(i) in [self.ns, 0] for i in [self.eqm]) and len(self.eqm) == self.ns:
                break
        self.ns = len(self.eqm)
        # Turn everything into arrays, convert to kcal/mol
        self.eqm = np.array(self.eqm)
        self.eqm *= (eqcgmx / 4.184)

    def indicate(self):
        delta = (self.emm-self.eqm)
        deltanrm = self.prefactor*(delta/self.divisor)**2
        if len(self.label) == self.ns:
            PrintDict = OrderedDict()
            for i,label in enumerate(self.label):
                PrintDict[label] = "% 9.3f % 9.3f % 9.3f % 9.3f % 11.5f" % (self.emm[i], self.eqm[i], delta[i], self.divisor[i], deltanrm[i])
            printcool_dictionary(PrintDict,title="Target: %s\nInteraction Energies (kcal/mol), Objective = % .5e\n %-10s %9s %9s %9s %9s %11s" %
                                 (self.name, self.objective, "Label", "Calc.", "Ref.", "Delta", "Divisor", "Term"),keywidth=15)
        else:
            # logger.info("Target: %s Objective: % .5e (add LABEL keywords in qdata.txt for full printout)\n" % (self.name,self.objective))
            Headings = ["Observable", "Difference\nRMS (Calc-Ref)", "Denominator\n(Specified)", " Percent \nDifference"]
            Data = OrderedDict([])
            Data['Energy (kcal/mol)'] = ["%8.4f" % np.sqrt(np.mean(delta**2)),
                                       "%8.4f" % np.mean(self.divisor),
                                       "%.4f%%" % (np.sqrt(np.mean(delta/self.divisor)**2)*100)]
            self.printcool_table(data=Data, headings=Headings, color=0)
            logger.info("add LABEL keywords in qdata.txt to print out each snapshot\n")


        # if len(self.RMSDDict) > 0:x
        #     printcool_dictionary(self.RMSDDict,title="Geometry Optimized Systems (Angstrom), Objective = %.5e\n %-38s %11s %11s" % (self.rmsd_part, "System", "RMSD", "Term"), keywidth=45)

    def get(self, mvals, AGrad=False, AHess=False):
        """ Evaluate objective function. """
        Answer = {'X':0.0, 'G':np.zeros(self.FF.np), 'H':np.zeros((self.FF.np, self.FF.np))}

        # If the weight is zero, turn all derivatives off.
        if (self.weight == 0.0):
            AGrad = False
            AHess = False

        def callM(mvals_, dielectric=False):
            logger.info("\r")
            pvals = self.FF.make(mvals_)
            return self.engine.interaction_energy(self.select1, self.select2)

        logger.info("Executing\r")
        emm = callM(mvals)

        D = emm - self.eqm
        dV = np.zeros((self.FF.np,len(emm)))

        if self.writelevel > 0:
            # Dump interaction energies to disk.
            np.savetxt('M.txt',emm)
            np.savetxt('Q.txt',self.eqm)
            import pickle
            pickle.dump((self.name, self.label, self.prefactor, self.eqm, emm), open("qm_vs_mm.p",'w'))
            # select the qm and mm data that has >0 weight to plot
            qm_data, mm_data = [], []
            for i in range(len(self.eqm)):
                if self.prefactor[i] != 0:
                    qm_data.append(self.eqm[i])
                    mm_data.append(emm[i])
            plot_interaction_qm_vs_mm(qm_data, mm_data, title="Interaction Energy "+self.name)

        # Do the finite difference derivative.
        if AGrad or AHess:
            for p in self.pgrad:
                dV[p,:], _ = f12d3p(fdwrap(callM, mvals, p), h = self.h, f0 = emm)
            # Create the force field one last time.
            pvals  = self.FF.make(mvals)

        Answer['X'] = np.dot(self.prefactor*D/self.divisor,D/self.divisor)
        for p in self.pgrad:
            Answer['G'][p] = 2*np.dot(self.prefactor*D/self.divisor, dV[p,:]/self.divisor)
            for q in self.pgrad:
                Answer['H'][p,q] = 2*np.dot(self.prefactor*dV[p,:]/self.divisor, dV[q,:]/self.divisor)

        if not in_fd():
            self.emm = emm
            self.objective = Answer['X']

        ## QYD: try to clean up OpenMM engine.simulation objects to free up GPU memory
        try:
            if self.engine.name == 'openmm':
                if hasattr(self.engine, 'simulation'): del self.engine.simulation
                if hasattr(self.engine, 'A'): del self.engine.A
                if hasattr(self.engine, 'B'): del self.engine.B
        except:
            pass

        return Answer

def plot_interaction_qm_vs_mm(eqm, emm, title=''):
    import matplotlib.pyplot as plt
    plt.plot(eqm, label='QM Data', marker='^')
    plt.plot(emm, label='MM Data', marker='o')
    plt.legend()
    plt.xlabel('Snapshots')
    plt.ylabel('Interaction Energy (kcal/mol)')
    plt.title(title)
    plt.savefig("e_qm_vs_mm.pdf")
    plt.close()



#!/usr/bin/env python

from __future__ import division
from __future__ import print_function
from builtins import map
from builtins import input
from builtins import range
from numpy import *
import os
from sys import argv,exit
import forcebalance
from forcebalance.PT import PeriodicTable
from forcebalance.nifty import isint, printcool
import getopt

opts,args = getopt.getopt(argv[1:],'h')

AutoH = False
for o,a in opts:
    argv.remove(o)
    if o == "-h":
        AutoH = True

Alphabet = list(map(chr, range(65, 91)))

printcool(" Welcome to Lee-Ping's Force Field Generator --- ")
print(" This script attempts to generate an OPLS force field ")
print(" from any Gaussian input file.  Here's how it works.  ")
print(" 1.  Define your atom types.                          ")
print(" 2.  Define corresponding OPLS atom types.            ")
print("     If there is no corresponding OPLS type, then you ")
print("     must define your own.")
print(" 3.  Select or input bonding parameters.              ")
print("     If you selected OPLS atom types, then the program")
print("     will attempt to recommend parameters for you.    ")
print(" 4.  Finished: GRO and ITP files will be generated.   ")

"""
gauss2gro-OPLS.py is an automatic force field maker, which builds a
force field from a molecular topology (in Gaussian .com format) and
fills in parameters automatically from OPLS-AA.  This is probably the
most complicated script I've written next to ForceBalance.  To avoid
confusion, this script DOES NOT do parameter optimization, it just
generates a force field from the OPLS parameter libraries for pretty
much any molecule you want.  You still have to fill in parameters by
hand for any gaps in OPLS (which are plenty).
 
Here's how it works.

As an input, we must have a Gaussian .com file with connectivity.
Usually, it's easiest to open a .com file with GaussView and resave
it, making sure that geom=connectivity is in the first line.

To enable automated atom-type definition (very helpful if you want to
follow OPLS 'standards',) you need to do either of the following:

OPTION 1 (recommended)
In the Gaussian .com file, after every atom name, type in '!opls_xxx' 
where opls_xxx is the standard opls atomtype 
(look it up using 'explain-opls.py' C for carbon).

OPTION 2:
In the Gaussian .com file, after every atom name, type in '!atom_type' 
where atom_type is an atom type of your choosing.
Somewhere else in the Gaussian file (preferably after the first line
and before the xyz coordinates), type in '! atom_type opls_xxx' to
identify that atom type with the OPLS parameters.

When the program is run, the first thing it does is try to get the
OPLS atom types.  You can assign atom types using one of the two
options above, in which case it will get the whole list automatically.
Otherwise, you can define atom types by element, or more specifically
by element + nearest neighbors.  In the latter case, you will be asked
to pick OPLS atom types manually, or enter your own nonbonded
parameters.  Picking OPLS atom types has the advantage that the bonded
parameters will be automatically chosen for you.

Once the atom types are chosen, the program will sum up all of the
atomic charges.  Because a simulation generally likes to have an
integer (usually zero) total charge, you will be given the option to
add a constant charge to every atom in order to make the overall
molecule neutral.  Finally, there is a fine-tuning step which allows
you to add small amounts to the charge so that it's _exactly_ zero,
despite the finite floating-point precision of the force field file.

The next thing the program does is to automatically parameterize all
of the detected bonds, angles, and dihedral types.  To do this, there
is a rather complicated set of lists.  Basically, in GROMACS there
each atom has several types associated with it: the 'element' (e.g. C
= 6), the 'base type', the 'OPLS type', the 'bonded type', and the
'OPLS nickname' defined in ffoplsaanb.itp.  In orthodox OPLS, the base
type starts with 'opls_', and is the same as the OPLS type; the bonded
atom type (e.g. 'CA', 'C_2' etc). and the OPLS nickname are the same.
In custom force fields, the base type, and the bonded atom type are
the same, but the OPLS type and the OPLS nickname are different.  The
OPLS type and nickname are essential for pulling the bonded parameters
out of the ffoplsaabon.itp file.  Thus, all four lists are important.

awkx gives a list of atoms and Cartesian coordinates, much like the
data in the xyz file.  awkc gives a list of the bonds at the end of
the Gaussian .com file.  awkpd gives a list of OPLS atom types which
correspond to the custom atom types, useful only in OPTION 2 above.
awkat gives a N-length list of cutom atom types or OPLS atom types,
depending on what the user has defined.

Below are explanations of names of variables.

oplsatoms, oplsnb, oplsbon are OPLS force field files, which are
essential to the working of this script.  They MUST point to the
locations of these files for automatic parameterization.
OPLSExplanations is a dictionary of strings from ffoplsaa.atp of 'OPLS
type':
E is a N-length list of elements.
R is a 3N-length array of Cartesian coordinates.
T is a N-length list of base types.
sbonds is a N-length list of lists, where the i-th entry contains all
 of the atoms j bonded to atom i.  It's the same as abonds (from the
 Gaussian file), except that bonds are bidirectional (i.e. the j-h
 entry still lists i as a bonded atom, even if j>i).
M is a dictionary of 'base type' : [atomic number, atomic mass, sigma,
 epsilon].  As you can see, each base type has its own nonbonded
 parameters.
BAT is a dictionary of 'base type': 'bonded type'
TEDict is a dictionary of 'base type U bonded type': 'element'
OPLSNicks is a dictionary of 'bonded type': 'OPLS nickname'
BList, AList, and DList are lists of interactions, which are built by
 searching the topology.  They are printed out in the later part of the
 force field file which lists all of the specific bonded interactions.
BTypeList, ATypeList, and DTypeList are lists of interaction types,
 which are built by searching the topology and enumerating the unique
 combinations of bonds, angles, and dihedrals (in terms of the OPLS
 nicknames).  They enter into the DefineParameters function, which
 returns dictionaries of interaction types ready for printing.
BondTypes, AngleTypes, and DihedralTypes are dictionaries of
 'bonded_type_1.bonded_type_2...' : [InteractionClass,[Bonded
 Parameters]].  These dictionaries are all created in the
 'DefineParameters' function, all enter into the 'PrintTypeSection'
 function as the 'SectionDict' argument.  OPLS nicknames are separated
 with periods, because OPLS nicknames themselves do not contain
 periods.  For example, a key in DihedralTypes would look like
 'CT.CA_2.CB.C!' and the value would look like [[3],[0.0 1.0 2.0 3.0
 4.0 5.0]].
AtomTypeCount is a dictionary of 'base type' : 'number of
 occurrences', which is used for zeroing out the net charge.
QDict is a dictionary of 'base type': 'atomic charge'.  Note that the
 atomic charge is printed out in the atoms section only; currently I
 choose to not print out any charges in the atomtypes section, as they
 are defined in the atoms section anyway..
predict and preatom are used for generating pre-defined atom types.
TDict is only used for generating neighbor-based atom types.

As of October 2012, this script is a part of ForceBalance and
covered under the same copyright notice and license.
"""

# Gets the molecule name
molname = argv[1][:3].upper()
molnametemp = input("Enter molecule name (three letters, default %s) >> " % molname)
if len(molnametemp) == 3:
    molname = molnametemp
else:
    print("Going with the default name (%s)" % molname)

# ENHANCED, SUPER-SPECIFIC FILTERS!!!
awkx = "awk -F '!' '{print $1}' %s | awk \'(NF==4 && $1*1!=$1 && $2*1==$2 && $3*1==$3 && $4*1==$4){print}\' " % argv[1]
awkc = "awk -F '!' '{print $1}' %s | awk \'($1==sprintf(\"%%i\",$1) && NF%%2==1){p=1;for(i=1;i<=NF;i++){if(i%%2==0){if($i!=sprintf(\"%%i\",$i)){p=0}}else{if($i!=$i*1){p=0}}}if(p==1){print}}\'" % argv[1]
awkpd = "grep '!' %s | awk '/ATYPE/ {print $(NF-1),$NF}'" % argv[1]
awkat = "awk \'(NF>=4 && $1*1!=$1 && $2*1==$2 && $3*1==$3 && $4*1==$4){print}\' %s | awk -F '!' '{if (NF>1) {print $NF}}' | awk '{print $1}'" % argv[1]

# Locations of files
datadir = os.path.split(forcebalance.__file__)[0]
oplsatoms = os.path.join(datadir,'data','oplsaa.ff','atomtypes.atp')
oplsnb = os.path.join(datadir,'data','oplsaa.ff','ffnonbonded.itp')
oplsbon = os.path.join(datadir,'data','oplsaa.ff','ffbonded.itp')

OPLSExplanations = {}
for line in os.popen("awk '$1 !~ /;/' %s" % oplsatoms).readlines():
    doc = line.split(";")[1].strip()
    OPLSExplanations[line.split()[0]] = doc

# Gets elements and cartesian coordinates
E = [l.split()[0] for l in os.popen(awkx)] # Element list
R = array([[float(i) for i in l.split()[1:]] for l in os.popen(awkx)])/10 # Cartesian coordinate list

# Gets list of bonds
abonds = [sorted([int(i)-1 for i in l.split()[1::2]]) for l in os.popen(awkc)]

# Gets lists of pre-defined atom types
predict = {}
for l in os.popen(awkpd):
    predict[l.split()[0]] = l.split()[1]
preatom = [l.strip() for l in os.popen(awkat)]
for l in preatom:
    if "opls_" in l:
        predict[l] = l

# Initialize variables
sbonds = []
T = []
BAT = {}
TDict = {} # BONDS: TYPE
TEDict = {} # TYPE: ELEMENT
na = len(abonds)
M = {}
OPLSNicks = {}
AtomTypeCount = {}
BList = [] # List of Bonds
AList = [] # List of Angles
DList = [] # List of Dihedrals
BTypeList = []
ATypeList = []
DTypeList = []
QDict = {} # List of Charges

# Alternate angle parameters that kick in if the angle exceeds ThreshAngle.  This part is ugly and might need rewriting.
altangles = {"O.Co.O":[[1],[1.4661e+02,5.4081e+01]],
             "O.Co.OA":[[1],[1.4661e+02,5.4081e+01]],
             "O.Co.OB":[[1],[1.3392e+02,5.1575e+01]],
             "O.Co.OC":[[1],[1.4661e+02,5.4081e+01]],
             "OA.Co.OA":[[1],[1.4661e+02,5.4081e+01]],
             "OA.Co.OB":[[1],[1.4661e+02,5.4081e+01]],
             "OA.Co.OC":[[1],[1.4661e+02,5.4081e+01]],
             "OB.Co.OB":[[1],[1.4661e+02,5.4081e+01]],
             "OB.Co.OC":[[1],[1.4661e+02,5.4081e+01]],
             "OC.Co.OC":[[1],[1.4661e+02,5.4081e+01]],
             "Co.OB.Co":[[1],[1.4661e+02,1.0312e+03]],
             "Co.OC.Co":[[1],[1.4661e+02,1.0312e+03]]}
ThreshAngle = 150

### FUNCTION DEFINITIONS ###

# In: An atom number and all of the atoms bonded to it.
# Out: All of the angle interactions with this atom at the center.
def acomb(a,set):
    c = []
    for i in set:
        for j in set:
            if j>i:
                c.append([i,a,j])
    return c

# In: Atom number A, atom number B, the atoms bonded to atom A, the atoms bonded to atom B.
# Out: All of the dihedral interactions for atom A and atom B.
def dcomb(a,b,seta,setb):
    d = []
    for i in seta:
        for j in setb:
            if i != b and j != a:
                if j>i:
                    d.append([i,a,b,j])
                elif j<i:
                    d.append([j,b,a,i])
    return d

# In: Atom number A, atom number B
# Out: The Cartesian distance between the two atoms in nanometers.
def dist(a,b):
    return dot(R[b]-R[a],R[b]-R[a])**0.5

# In: Three atom numbers
# Out: The angle ABC in degrees.
def angle(a,b,c):
    ab = R[a]-R[b];
    cb = R[c]-R[b];
    theta=arccos(dot(ab,cb)/(dot(ab,ab)**0.5*dot(cb,cb)**0.5))*180/pi
    return theta

# Recursive ring finder
# This is a cool function but we currently don't use it for anything
def ring(size,STEP,NOW,PREV):
    FWD = [i for i in sbonds[NOW]]
    for i in PREV[1:]:
        if i in FWD:
            FWD.remove(i)
    if STEP <= size:
        for i in FWD:
            ring(size,STEP+1,i,PREV+[NOW])
    if STEP == size+1 and PREV[-1] == PREV[0]:
        if sorted(PREV[:-1]) not in ring.rsort:
            ring.ratoms.append(PREV[:-1])
            ring.rsort.append(sorted(PREV[:-1]))

# Recursive base type definition
# The initial call is with atomtype(elem,elem)
# Afterward, it goes through the list of base types (CA, CB, CC, CD...)
# until it finds an unused name and returns it
def atomtype(E, A):
    if A in T:
        if A == E:
            A += "A"
            return atomtype(E, A)
        else:
            ANew = A[:-1] + Alphabet[Alphabet.index(A[-1])+1]
            return atomtype(E, ANew)
    else:
        return A

# Used in neighbor-based base type definition
# Returns a unique "key" corresponding to all of the bonded atoms of atom A.
def BuildAKey(i):
    AllNeighbors = [E[j] for j in sbonds[i]]
    AllNeighbors.sort()
    AKey = E[i] + "." + ".".join(AllNeighbors)
    return AllNeighbors, AKey

# This simply prints out the atom number, element, neighbors, and base type of a given atom.
def PrintInfo(i):
    AllNeighbors,AKey = BuildAKey(i)
    print("%9i%10s%16s%15s" % (i+1,E[i]," ".join(AllNeighbors),T[i]))

# NOTE: This function is only called by DefineParameters
# Input: Section name ('bondtypes','angletypes','dihedraltypes'), interaction type ('CA.CB.CT')
# Output: Filtered OPLS-AA parameters.
# This goes through the ffoplsaabon.itp file and filters out relevant lines for parameter selection.
# A special feature is "recommendations": If the OPLS nicknames for the input interaction type matches the nicknames on a given line,
# that line is given "points" by the number of matches, denoted by a red carat ^.  
# Note: If we're working on a dihedral interaction, then the two middle atoms get three points while the edge atoms get one point.
# If all of the nicknames match, then that line gets ten points (otherwise impossible).
# Also: "X" is a wild card for dihedral interactions; it counts as a match but doesn't qualify for ten points.
def OPLS_Filter(SectionName,Type):
    ElementList = [TEDict[i] for i in Type.split(".")]
    NickList = [OPLSNicks[i] for i in Type.split(".")]
    rawfile = os.popen("awk '/%s/,(NF==0)' %s" % (SectionName,oplsbon)).readlines()
    filtered = []
    original = []
    already = []
    elemrev = ElementList[::-1]
    typerev = NickList[::-1]
    for line in rawfile:
        sline = line.split()
        try:
            if len(NickList) == 4:
                score = max(sum(array([NickList[i] == sline[i] for i in range(len(NickList))])*array([1,3,3,1])),sum(array([typerev[i] == sline[i] for i in range(len(NickList))])*array([1,3,3,1]))) + line.count(" X ")
            else:
                score = max(sum(array([NickList[i] == sline[i] for i in range(len(NickList))])),sum(array([typerev[i] == sline[i] for i in range(len(NickList))]))) + line.count(" X ")
            if line in already: continue
            already.append(line)
            if array([NickList[i] == sline[i] for i in range(len(NickList))]).all():
                filtered.append(line.replace("\n","")+"\x1b[91m^^^^^^^^^^\x1b[0m\n")
                original.append(line)
            elif array([typerev[i] == sline[i] for i in range(len(NickList))]).all():
                filtered.append(line.replace("\n","")+"\x1b[91m^^^^^^^^^^\x1b[0m\n")
                original.append(line)
            elif array([ElementList[i] in sline[i][:len(ElementList[i])] or sline[i] == "X" for i in range(len(ElementList))]).all():
                if score > 0:
                    filtered.append(line.replace("\n","")+"\x1b[91m%s\x1b[0m\n" % (''.join(["^" for i in range(score)])))
                    original.append(line)
                else:
                    filtered.append(line)
                    original.append(line)
            elif array([elemrev[i] in sline[i][:len(elemrev[i])] or sline[i] == "X" for i in range(len(elemrev))]).all():
                if score > 0:
                    filtered.append(line.replace("\n","")+"\x1b[91m%s\x1b[0m\n" % (''.join(["^" for i in range(score)])))
                    original.append(line)
                else:
                    filtered.append(line)
                    original.append(line)
        except: pass
    return filtered, original

# Default interaction for when there are no input parameters.
# Usually returns all zeros, but also gives default improper
# dihedrals for aromatic rings.
def getdefaults(Type,SectionName,Positions):
    Answer = zeros(len(Positions))
    Comment = 'Default: all zeros'
    if SectionName == 'dihedral':
        Answer[0] = 30.33400
        Answer[2] = -30.33400
        Comment = 'Default: aromatic ring'
    return Answer, Comment

# Takes in the text from OPLS_Filter and gives back recommendations.
# This is very simple; it just gives back the line with the highest number of points.
# If the automatic switch is turned on, then it will choose this line automatically.
# If there is no line with ten points, then the answer is ambiguous;
# it will then prompt the user to choose (the user isn't restricted to choosing
# the most recommended one.)
def recommend(textlist,manual):
    goodlines = []
    currcount = 0
    for line in textlist:
        starcount = line.count("^")
        if starcount > currcount:
            currcount = starcount
            goodlines = [textlist.index(line)]
        elif starcount == currcount:
            goodlines.append(textlist.index(line))
    if len(goodlines) == 1 and currcount == 10:
        return manual and 1 or 0, goodlines
    elif len(goodlines) >= 1:
        return 1, goodlines
    else:
        return 1, goodlines

# In: A list of atom types separated by periods, as for an interaction type
# Out: The ordered list of atom types
# This function takes care to swap atom types in dihedrals where A and D are the same,
# but B and C are not necessarily the same.
def Order(string):
    ssd = string.split(".")
    if ssd[0] > ssd[-1]:
        newlist = ssd[::-1]
    else:
        if len(ssd) == 4:
            if ssd[0] == ssd[3] and ssd[1] > ssd[2]:
                newlist = [ssd[0],ssd[2],ssd[1],ssd[3]]
            else:
                newlist = ssd
        else:
            newlist = ssd
    return ".".join(newlist)

# In: SectionName (bondtypes, angletypes, dihedraltypes) and SectionDict (from DefineParameters)
# and prints out a formatted block of parameters, either to the itp file or to screen.
def PrintTypeSection(SectionName,SectionDict):
    answer = []
    answer.append("[ %s ]" % SectionName)
    for i in sorted(SectionDict):
        line = ""
        for j in i.split("."):
            line += "%5s" % j
        line += "%5i" % SectionDict[i][0]
        for j in SectionDict[i][1]:
            line += "% 14.4e" % j
        if len(SectionDict[i][2]) > 0:
            line += " ; %s" % SectionDict[i][2]
        answer.append(line)
    answer.append("")
    return answer

# In: SectionName, list of 2 atoms (bonds) or 3 atoms (angles)
# Out: Either the distance between 2 atoms or the angles between 3 atoms
def GetBaseVals(SectionName,Atoms):
    if SectionName == "bonds":
        return "%10.5f" % dist(Atoms[0],Atoms[1])
    elif SectionName == "angles":
        return "%12.5f" % angle(Atoms[0],Atoms[1],Atoms[2])
    else:
        return ""

# Prints the 1-4 interactions using the dihedral list.
def PrintPairs(blist,dlist):
    answer = ["[ pairs ]"]
    printed = []
    for i in sorted(dlist):
        if sorted([i[0],i[-1]]) not in blist and sorted([i[0],i[-1]]) not in printed:
            printed.append(sorted([i[0],i[-1]]))
    for i in sorted(printed):
        line = "%5i%5i" % (i[0]+1,i[-1]+1)
        answer.append(line)
    answer.append("")
    return answer

# In: SectionName (bonds, angles, dihedrals), SectionDict (from DefineParameters),
# SectionList (a list of the specific interactions), PrintBaseVals (a switch),
# Sub180 (substitute parameters for large angles)
# In this section, the specific interaction is printed out, along with the
# base value (bond length, angle) and force constant if desired.
def PrintItemSection(SectionName,SectionDict,SectionList,PrintBaseVals,Sub180):
    answer = []
    answer.append("[ %s ]" % SectionName)
    for i in sorted(SectionList):
        line = ""
        for j in i:
            line += "%5i" % (j+1)
        line += "%5i" % SectionDict[Order(".".join([BAT[T[j]] for j in i]))][0]
        if SectionName == "angles" and Sub180 and float(GetBaseVals(SectionName,i)) > ThreshAngle:
            try:
                line += "% 14.4e" % altangles[Order(".".join([BAT[T[j]] for j in i]))][1][-2]
                line += "% 14.4e" % altangles[Order(".".join([BAT[T[j]] for j in i]))][1][-1]
            except:
                print("Sub180 has failed for angle %s" % (".".join(["%i" % j for j in i])))
        elif PrintBaseVals:
            line += GetBaseVals(SectionName,i)
            line += "% 14.4e" % SectionDict[Order(".".join([BAT[T[j]] for j in i]))][1][-1]
        answer.append(line)
    answer.append("")
    return answer

# This is the big lunker!
# In: InteractionList (a list of the specific interaction), TypeList (list of the bonded interaction types 'CA.CB.CT'), 
# InteractionName ('bondtypes','angletypes','dihedraltypes'), InteractionClass (1 for bonds, 1 for angles, 3 for dihedrals),
# Positions (integers corresponding to fields in line where the parameters belong)
# First ask the user whether to go into Automatic Mode
# For each interaction type (e.g. 'CA.CB.CT'), do the following:
# Get a selection of candidate lines from ffoplsaabon.itp and recommend parameters
# Pick the parameter automatically if in automatic mode and an unambiguous choice is made
# Ask the user if there are any decisions to be made
# When we are all done, give back a dictionary with InteractionTypes:Parameters
def DefineParameters(InteractionList,TypeList,InteractionName,SectionName,InteractionClass,Positions):
    SectionDict = {}
    choice = input("%i '%s' interactions of %i types have been found.  Enter 'Yes' to perform automatic parameterization. --> " % (len(InteractionList),InteractionName,len(TypeList)))
    if InteractionName == 'bonds' or InteractionName == 'angles':
        ReadBaseVals = input("Enter 'Yes' to get equilibrium values from source coordinate file (Useful if you did a geometry optimization!) -->") + " "
    if len(choice) > 0:
        if 'y' == choice[0] or 'Y' == choice[0]:
            Manual,ManualNow = 0,0
        else:
            Manual,ManualNow = 1,1
    else:
        Manual,ManualNow = 0,0
    try:
        if 'y' == ReadBaseVals[0] or 'Y' == ReadBaseVals[0] or ReadBaseVals == " ":
            ReadBaseVals = 1
        else:
            ReadBaseVals = 0
    except:
        ReadBaseVals = 0
    for Type in TypeList:
        Good = []
        TypeSplit = Type.split(".")
        Comment = 'OPLS'
        DefaultComment = 'OPLS'
        if ReadBaseVals:
            BaseVals = []
            for item in InteractionList:
                # This complicated line gets a list of base bond-lengths or angles from the atoms that match
                # Note that the data structures are apparent here: OPLSNicks[Bonded_Type[Base_Type[Atom_Number]]]
                if [OPLSNicks[BAT[T[i]]] for i in item] == [OPLSNicks[i] for i in TypeSplit] or [OPLSNicks[BAT[T[i]]] for i in item] == [OPLSNicks[i] for i in TypeSplit[::-1]]:
                    BaseVals.append(float(GetBaseVals(InteractionName,item)))
            BaseVals = array(BaseVals)
        OPLSChoices, RawLines = OPLS_Filter(SectionName,Type)
        ManualNow, Good = recommend(OPLSChoices,Manual)
        if not Manual:
            if ManualNow:
                print("Automatic parameterization failed (but don't worry), going to Manual Selection")
        if not ManualNow:
            Selection = Good[0]
            Parameters = [float(OPLSChoices[Selection].split()[i]) for i in Positions]
            if len(RawLines[Selection].split(';')) > 1:
                DefaultComment = "OPLS " + RawLines[Selection].split(';')[1].strip()
            if ReadBaseVals:
                Parameters[0] = mean(BaseVals)
                print("Automatic parameterization of %s %s using base geometry, std deviation is %.6f" % (Type,InteractionName,std(BaseVals)))
                DefaultComment += ', geom from xyz'
            print("We got these parameters:", end=' ')
            print("".join(["%-6s" % i for i in Type.split(".")]), "%5i" % InteractionClass, "".join(["% 10.5f" % i for i in Parameters]))
            print()
        else:
            DefaultParams, DefaultComment = getdefaults(Type,InteractionName,Positions) # getdefaults code returns zeros most of the time.
            try:
                try:
                    DefaultParams = [float(OPLSChoices[Good[0]].split()[i]) for i in Positions]
                    if len(RawLines[Good[0]].split(';')) > 1:
                        DefaultComment = "OPLS " + RawLines[Good[0]].split(';')[1].strip()
                except:
                    DefaultParams = [float(OPLSChoices[0].split()[i]) for i in Positions]
                    if len(RawLines[0].split(';')) > 1:
                        DefaultComment = "OPLS " + RawLines[0].split(';')[1].strip()
            except:
                pass
            if ReadBaseVals:
                DefaultParams[0] = mean(BaseVals)
                DefaultComment += ', geom from xyz'
            for line in OPLSChoices:
                print("%5i" % OPLSChoices.index(line),line, end=' ')
            if len(Good) == 0:
                print(" No recommended choices! ")
            else:
                print(" --- RECOMMENDED CHOICES --- ")
                for item in Good:
                    print(item, OPLSChoices[item], end=' ')
            print(" --- DEFAULT PARAMETERS  ---")
            print("".join(["%-6s" % i for i in Type.split(".")]), "%5i" % InteractionClass, "".join(["% 10.5f" % i for i in DefaultParams]))
            if ReadBaseVals:
                print("Standard deviation of all equilibrium %s of this type is %.6f" % (InteractionName,std(BaseVals)))
            line = input("\nFor a '%s' interaction involving %s (OPLS types %s), \nselect line number from above, type your own parameters, \nor hit Enter to accept default. (# of \x1b[91m^\x1b[0m symbols indicate score) --> " % (InteractionName,"-".join(Type.split(".")),"-".join([OPLSNicks[i] for i in Type.split(".")])))
            try:
                Selection = int(line.strip())
                Parameters = [float(OPLSChoices[Selection].split()[i]) for i in Positions]
                print("Going with the following parameter selection: ")
                print(Selection, OPLSChoices[Selection])
                if len(RawLines[Selection].split(';')) > 1:
                    Comment = "OPLS " + RawLines[Selection].split(';')[1].strip()
            except:
                try:
                    if len(line.split()) == len(Positions):
                        print("Going with user-specified parameters.")
                        Parameters = [float(i) for i in line.split()]
                        Comment = 'User Specified Parameter'
                    else:
                        print("No selection or user-specified parameters given, going with the default parameters.")
                        Parameters = DefaultParams
                        Comment = DefaultComment
                except:
                    print("No selection or user-specified parameters given, going with the default parameters.")
                    Parameters = DefaultParams
                    Comment = DefaultComment
            print()
        SectionDict[Type] = [InteractionClass,Parameters,Comment]
    print("--- Finished Selecting %s Parameters ---" % InteractionName)
    for line in PrintTypeSection(SectionName,SectionDict):
        print(line)
    return SectionDict

### END FUNCTION DEFINITIONS ###

### BUILDING CONNECTIVITY MATRIX ###
        
for i in range(na):
    sbvec = [k for k in abonds[i]]
    for j in range(na):
        if i in abonds[j]:
            sbvec.append(j)
    sbonds.append([k for k in sort(sbvec)])

### DEFINING ATOM TYPES ###

if len(preatom) == len(sbonds):
    print("Using pre-defined atom types in Gaussian .com file (syntax Element X Y Z ! AtomType)")
    AtomTypeSwitch = 1
else:
    AtomTypeSwitch = input("Enter '1' for neighbor-based atom type generation or '0' for simple element-based atom types. -->")[0] == '1'

print("\n--- Number --- Element --- Neighbors --- Assigned Type ---\n")
for i in range(len(sbonds)): # [ELEMENT NEIGHBOR1 NEIGHBOR2...]
    AllNeighbors,AKey = BuildAKey(i)
    if len(preatom) == len(sbonds):
        AtomType = preatom[i]
    else:
        try:
            AtomType = TDict[AKey]
        except:
            AtomType = atomtype(E[i],E[i])
            TDict[AKey] = AtomType
    # LPW added a switch to make all carbon-bonded hydrogen "HC" and all oxygen-bonded hydrogen "HO"
    if AtomTypeSwitch:
        T.append(AtomType)
    else:
        T.append(E[i])
    PrintInfo(i)
    TEDict[AtomType] = E[i]

choice = input("Modify the list? (Enter for no) --> ") + " "
print()
while 'y' == choice[0] or 'Y' == choice[0]:
    print("\n--- Number --- Element --- Neighbors --- Assigned Type ---\n")
    for i in range(len(T)):
        PrintInfo(i)
    line = input("Enter the index of the atom you wish to change, or enter Q to quit. -->")
    print()
    if line == "Q" or line == 'q':
        break
    Index = int(line.strip())
    line = input("Enter the new type of atom %i. --> " % Index)
    print()
    Type = line.strip()
    try:
        T[Index-1] = Type
        TEDict[Type] = E[Index-1]
    except:
        choice = input("There was an error.  Do you want to continue? --> ")
        print()

input("\nAtom Types Saved. Press Enter")

### SEARCHING FOR RINGS ###

ring.ratoms = []
ring.rsort = []

print("\nFinding Rings...", end=' ')
for i in range(na):
    ring(6,0,i,[])
    ring(5,0,i,[])
for i in ring.ratoms:
    print("Ring found at",i)
print("...Done")

### SELECTING NONBONDED PARAMETERS ###

# This section should probably be its own function, but it's only called once.
# This loop goes through the base atom types and selects the bonded types and OPLS nicknames.
# If the types are not chosen automatically, then the user gets to pick from a list.
print("%i atoms of %i distinct types have been found.  Press Enter." % (len(T),len(set(T))))
input()
for i in sorted(list(set(T))):
    Element = TEDict[i]
    OPLSChoices = list(os.popen("awk '$2 ~ /%s/' %s | awk '(($4 - %.3f) < 0.1 && ($4 - %.3f) > -0.1)'" % (Element,oplsnb,PeriodicTable[Element],PeriodicTable[Element])).readlines())
    ChoicesByName = {}
    for line in OPLSChoices:
        ChoicesByName[line.split()[0]] = line
    try:
        # The contents of this if statement are invoked only if a valid OPLS type is indicated in the Gaussian .com file
        # using either ! ATYPE custom_type opls_xxx or Element x y z ! opls_xxx
        if predict[i] in ChoicesByName:
            print("ATYPE definition found for %s!" % i)
            Selection = predict[i]
            SelectedLine = ChoicesByName[Selection]
            sline = SelectedLine.split()
            print("%s -> %10s%5s%5i%10.5f%10.3f%15.5e%15.5e  " % (i, sline[0],sline[1],int(sline[2]),float(sline[3]),float(sline[4]),float(sline[6]),float(sline[7])), OPLSExplanations[sline[0]])
            print()
            ANum,Q,SIG,EPS = int(SelectedLine.split()[2]),float(SelectedLine.split()[4]),float(SelectedLine.split()[6]),float(SelectedLine.split()[7])
            COM = "OPLS " + OPLSExplanations[sline[0]]
            # "Orthodox" OPLS; base types are opls_xxx.  In this case, go to new bonded atom types.
            if "opls" in i:
                BAT[i] = SelectedLine.split()[1]#OPLSNicks[i]
                OPLSNicks[BAT[i]] = BAT[i]
            # Base types and bonded types are custom names; don't have to use opls_xxx, but still assign OPLS nicknames
            else:
                BAT[i] = i
                OPLSNicks[BAT[i]] = SelectedLine.split()[1]
            TEDict[BAT[i]] = Element
            M[i] = [ANum,PeriodicTable[Element],SIG,EPS,COM]
            QDict[i] = float("%.4f" % Q)
            continue
    except: pass
    for line in OPLSChoices:
        sline = line.split()
        print("-> %5i <- %10s%5s%5i%10.5f%10.3f%15.5e%15.5e  " % (OPLSChoices.index(line),sline[0],sline[1],int(sline[2]),float(sline[3]),float(sline[4]),float(sline[6]),float(sline[7])), OPLSExplanations[line.split()[0]])
    line = input("For Atomtype %s, please select the corresponding OPLS atomtype from above by NUMBER (sequential) or NAME (opls_xxx), or enter three parameters of your own: Q SIG EPS --> " % i)
    print()
    if len(line.split()) == 3:
        try:
            # If the user enters custom parameters, then there is no corresponding OPLS atomtype.
            # In this case, there is no OPLS nickname either, and we will attempt to get one by simply setting it equal to the base type.
            Q,SIG,EPS = [float(j) for j in line.split()]
            COM = 'User Specified Atomtype'
            BAT[i] = i
            OPLSNicks[BAT[i]] = i
            TEDict[BAT[i]] = Element
        except:
            print("An error has occurred in getting nonbonded parameters.")
    else:
        try:
            try:
                Selection = int(line.strip())
                SelectedLine = OPLSChoices[Selection]
            except:
                Selection = line.strip()
                SelectedLine = ChoicesByName[Selection]
        except:
            print("An error has occurred in getting nonbonded parameters.")
        print(Selection,SelectedLine)
        ANum,Q,SIG,EPS = int(SelectedLine.split()[2]),float(SelectedLine.split()[4]),float(SelectedLine.split()[6]),float(SelectedLine.split()[7])
        COM = "OPLS " + OPLSExplanations[sline[0]]
        BAT[i] = i
        OPLSNicks[BAT[i]] = SelectedLine.split()[1]
        TEDict[BAT[i]] = Element
    M[i] = [ANum,PeriodicTable[Element],SIG,EPS,COM]
    QDict[i] = float("%.4f" % Q)

# Try to zero out the atomic charge.

print("The following code aims to get an integer-charge molecule using precision 4 floating points.")
TQ = sum([QDict[T[i]] for i in range(na)])
print("All atom types have been selected; molecule has a net charge of %.4f" % TQ)
Corr1 = (float("%.4f" % (-1*TQ/na)))
choice = input("Create overall integer charge (enter integer), enter your own (enter float), or skip this step (hit Enter)? --> ") + " "
#if 'y' == choice[0] or 'Y' == choice[0]:
if isint(choice.strip()):
    WantQ = int(choice)
    Corr1 = (float("%.4f" % ((WantQ - 1*TQ)/na)))
    for i in QDict:
        QDict[i] += Corr1
    TQ = sum([QDict[T[i]] for i in range(na)])
    print("Adding a charge of %.4f to each atom" % Corr1)
    print("Now all atom types have been selected; molecule has a net charge of %.4f" % TQ)
else:
    try:
        Corr1 = float(choice)
        for i in QDict:
            QDict[i] += Corr1
        TQ = sum([QDict[T[i]] for i in range(na)])
        print("Now all atom types have been selected; molecule has a net charge of %.4f" % TQ)
    except: print("You didn't enter a number, exiting this section")
                
for i in QDict:
    AtomTypeCount[i] = sum(i == array(T))

PrintSwitch = 0
while abs(TQ - int(TQ)) > 1e-5:
    print()
    print("Atom Counts:", AtomTypeCount)
    print("Total charge: % .4f" %  TQ)
    choice = input("Manually add charge to an atom type using: Atype dQ (Q or Enter quits) ? --> ") + " "
    if 'q' == choice[0] or 'Q' == choice[0] or choice == " ": break
    else:
        try:
            QDict[choice.split()[0]] += float(choice.split()[1])
        except: print("The input cannot be parsed")
    TQ = sum([QDict[T[i]] for i in range(na)])

print()
print("Charges are GOOD, molecule has a net charge of %.4f" % TQ)
print()

print("--- FINISHED SELECTING NONBONDED PARAMETERS ---")
for i in M:
    print(i, ": ANum = %5i M = %.4f Q = % .3f SIG = %.4f EPS = %.4f" % (M[i][0],M[i][1],QDict[i],M[i][2],M[i][3]))

### DETECTION AND PARAMETERIZATION OF BONDS ###

for i in range(na):
    for j in abonds[i]:
        BList.append(sorted([i,j]))
        BType = ".".join(sorted([BAT[T[i]],BAT[T[j]]]))
        if BType not in BTypeList:
            BTypeList.append(BType)

BTypeList.sort()
BondTypes = DefineParameters(BList,BTypeList,'bonds','bondtypes',1,[3,4])

### DETECTION AND PARAMETERIZATION OF ANGLES ###
    
for i in range(na):
    alist = acomb(i,sbonds[i])
    for j in alist:
        AList.append(j)
        acs = sorted([BAT[T[j[0]]],BAT[T[j[2]]]])
        AType = '.'.join([acs[0],BAT[T[j[1]]],acs[1]])
        if AType not in ATypeList:
            ATypeList.append(AType)

ATypeList.sort()
AngleTypes = DefineParameters(AList,ATypeList,'angles','angletypes',1,[4,5])

### DETECTION AND PARAMETERIZATION OF DIHEDRALS ###

for i in range(len(BList)):
    dlist = dcomb(BList[i][0],BList[i][1],sbonds[BList[i][0]],sbonds[BList[i][1]])
    for j in dlist:
        DList.append(j)
        if BAT[T[j[0]]] < BAT[T[j[3]]]:
            DType = '.'.join([BAT[T[j[i]]] for i in range(4)])
        elif BAT[T[j[0]]] > BAT[T[j[3]]]:
            DType = '.'.join([BAT[T[j[i]]] for i in range(3,-1,-1)])
        else:
            if BAT[T[j[1]]] <= BAT[T[j[2]]]:
                DType = '.'.join([BAT[T[j[i]]] for i in range(4)])
            else:
                DType = '.'.join([BAT[T[j[i]]] for i in range(3,-1,-1)])
        if DType not in DTypeList:
            DTypeList.append(DType)

DTypeList.sort()
DihedralTypes = DefineParameters(DList,DTypeList,'dihedral','dihedraltypes',3,[5,6,7,8,9,10])


def printitp():
    itpfile = open(argv[1].replace('.com','.itp'),'w')
    # Prints some header stuff.
    print('[ defaults ]', file=itpfile)
    print('%-5i%-5i%-10s%5.1f%5.1f' % (1,2,'yes',0.5,0.5), file=itpfile) # Defaults
    # Prints atom types and VdW parameters.
    print(file=itpfile)
    print('[ atomtypes ]', file=itpfile)
    for i in sorted([j for j in M]):
        if i == BAT[i]:
            print('%-10s%5i%10.4f%10.4f%5s%14.4e%14.4e ; %s' % (i,M[i][0],M[i][1],0.0,'A',M[i][2],M[i][3],M[i][4]), file=itpfile)
        else:
            Orthodox = 1
            print('%-10s%5s%5i%10.4f%10.4f%5s%14.4e%14.4e ; %s' % (i,BAT[i],M[i][0],M[i][1],0.0,'A',M[i][2],M[i][3],M[i][4]), file=itpfile)
    choice = input("Do you want to add atomtypes for a water model?? 0 for SPC(/E), 1 for TIP4P, anything else for No: -->")
    try:
        choice = int(choice)
        if choice == 0:
            print('%-10s%5s%5i%10.4f%10.4f%5s%14.4e%14.4e ; SPC/E Oxygen' % ("opls_116","OW",8,15.99940,-0.820,"A",3.16557e-01,6.50194e-01), file=itpfile)
            print('%-10s%5s%5i%10.4f%10.4f%5s%14.4e%14.4e ; SPC/E Hydrogen' % ("opls_117","HW",1,1.00800,0.410,"A",0.00000e+00,0.00000e+00), file=itpfile)
        elif choice == 1:
            print('%-10s%5s%5i%10.4f%10.4f%5s%14.4e%14.4e ; TIP4P Oxygen' % ("opls_113","OW",8,15.99940,0.000,"A",3.15365e-01,6.48520e-01), file=itpfile)
            print('%-10s%5s%5i%10.4f%10.4f%5s%14.4e%14.4e ; TIP4P Hydrogen' % ("opls_114","HW",1,1.00800,0.520,"A",0.00000e+00,0.00000e+00), file=itpfile)
            print('%-10s%5s%5i%10.4f%10.4f%5s%14.4e%14.4e ; TIP4P Virtual Site' % ("opls_115","MW",0,0.00000,-1.040,"D",0.00000e+00,0.00000e+00), file=itpfile)
        else:
            pass
    except: pass
    print(file=itpfile)
    # Prints parameter type definitions for bonds, angles, dihedrals.
    for line in PrintTypeSection("bondtypes",BondTypes):
        print(line, file=itpfile)
    for line in PrintTypeSection("angletypes",AngleTypes):
        print(line, file=itpfile)
    for line in PrintTypeSection("dihedraltypes",DihedralTypes):
        print(line, file=itpfile)
    # Prints moleculetype.
    print(file=itpfile)
    print('[ moleculetype ]', file=itpfile)
    print('%-10s%5i' % (molname,3), file=itpfile) # Number of exclusions
    print(file=itpfile)
    # Prints atoms.
    print('[ atoms ]', file=itpfile)
    SumCharges = 0.0
    for i in range(na):
        # atom = BAT[T[i]] + "%i" % (i+1)
        atom = E[i] + "%i" % (i+1)
        print('%5i%10s%5i%10s%7s%5i%10.4f%10.4f' % (i+1,T[i],1,molname,atom,i+1,QDict[T[i]],M[T[i]][1]), file=itpfile)
        SumCharges += QDict[T[i]]
    if abs(SumCharges) > 1e-6:
        print("; The total charge of the generated FF is %.4f." % SumCharges, file=itpfile)
    print(file=itpfile)
    # Prints atom lists for bonds, angles, dihedrals.
    choice = input("Print structural bond lengths into force field file? -->") + " "
    if 'y' == choice[0] or 'Y' == choice[0]:
        PrintBaseVals = 1
    else:
        PrintBaseVals = 0
    for line in PrintItemSection("bonds",BondTypes,BList,PrintBaseVals,0):
        print(line, file=itpfile)
    choice = input("Print structural angles into force field file? (Enter no for sub180.) -->") + " "
    Sub180 = 0
    if 'y' == choice[0] or 'Y' == choice[0]:
        PrintBaseVals = 1
    else:
        PrintBaseVals = 0
        choice = input("Try to substitute alternate parameters for linear angles (threshold is 150 degrees)? -->") + " "
        if 'y' == choice[0] or 'Y' == choice[0]:
            Sub180 = 1
    for line in PrintItemSection("angles",AngleTypes,AList,PrintBaseVals,Sub180):
        print(line, file=itpfile)
    PrintBaseVals = 0
    for line in PrintItemSection("dihedrals",DihedralTypes,DList,0,0):
        print(line, file=itpfile)
    choice = input("Print 1-4 pairs into force field file? (Seems the only way to make 1-4 interactions work.) -->") + " "
    if 'y' == choice[0] or 'Y' == choice[0]:
        for line in PrintPairs(BList,DList):
            print(line, file=itpfile)
    print("Parameter file written to %s" % (argv[1].replace('.com','.itp')))
    itpfile.close()

def printgro():
    grofile = open(argv[1].replace('.com','.gro'),'w')
    print('Generated by gauss2gro: %s' % molname, file=grofile)
    print(na, file=grofile)
    res = "1"+molname
    for i in range(na):
        #atom = BAT[T[i]] + "%i" % (i+1)
        atom = E[i] + "%i" % (i+1)
        print("%8s%7s%5i%12.7f%12.7f%12.7f" % (res,atom,i+1,R[i][0],R[i][1],R[i][2]), file=grofile)
    print("   3.00000   3.00000   3.00000", file=grofile)
    print("Coordinate file written to %s" % (argv[1].replace('.com','.gro')))
    grofile.close()
    
def main():
    printitp()
    printgro()
    print("Program Finished Successfully.  Have fun!")

main()



#!/usr/bin/env python

from builtins import range
import numpy as np
import os, sys
from forcebalance.nifty import isfloat
from forcebalance.molecule import Molecule

def read_psi_xyzesp(psiout):
    # Read Psi4 ESP output file for geometries, ESP values and grid points.
    XMode = 0
    EMode = 0
    ESPMode = 0
    xyzs = []
    xyz = []
    elem = []
    espxyz = []
    espval = []
    for line in open(psiout):
        s = line.split()
        if XMode == 1:
            if len(s) == 4 and isfloat(s[1]) and isfloat(s[2]) and isfloat(s[3]):
                e = s[0]
                xyz.append([float(i) for i in s[1:4]])
                if EMode == 1:
                    elem.append(e)
            elif len(xyz) > 0:
                xyzs.append(np.array(xyz))
                xyz = []
                XMode = 0
        if ESPMode == 1:
            if len(s) == 4 and isfloat(s[0]) and isfloat(s[1]) and isfloat(s[2]) and isfloat(s[3]):
                espxyz.append([float(i) for i in s[:3]])
                espval.append(float(s[3]))
            elif len(espxyz) > 0:
                # After reading in a block of ESPs, don't read any more.
                ESPMode = -1 
        if line.strip().startswith("Geometry (in Angstrom)"):
            XMode = 1
            EMode = len(elem) == 0
        if 'Electrostatic Potential' in line.strip() and ESPMode == 0:
            ESPMode = 1
    if len(xyzs) == 0:
        raise Exception('%s has length zero' % psiout)
    return xyzs, elem, espxyz, espval

xyzs, elem, espxyz, espval = read_psi_xyzesp(sys.argv[1])

M = Molecule()
M.xyzs = xyzs
M.elem = elem
M.write('%s.xyz' % os.path.splitext(sys.argv[1])[0])

EM = Molecule()
EM.xyzs = [np.array(espxyz) * 0.52917721092]
EM.elem = ['H' for i in range(len(espxyz))]
EM.write('%s.espx' % os.path.splitext(sys.argv[1])[0], ftype="xyz")

M.qm_espxyzs = EM.xyzs
M.qm_espvals = [np.array(espval)]
M.write("qdata.txt")

np.savetxt('%s.esp' % os.path.splitext(sys.argv[1])[0], espval)
